from __future__ import absolute_import
from __future__ import print_function
import os
import sys
import copy
from collections import defaultdict

import veriloggen.core.vtypes as vtypes

from . import stypes
from .visitor import _Visitor


def draw_graph(vars, filename='out.png', prog='dot', rankdir='LR', approx=False):
    gg = GraphGenerator(rankdir=rankdir, approx=approx)
    gg.draw(vars, filename, prog)


class GraphGenerator(_Visitor):

    def __init__(self, rankdir='LR', approx=False):
        _Visitor.__init__(self)

        try:
            import pygraphviz as pgv
        except:
            raise ImportError('Graph generator requires Pygraphviz.')

        self.graph = pgv.AGraph(directed=True, rankdir=rankdir)

        self.approx = approx
        self.tmp_count = 0

        self.ranks = defaultdict(list)
        self.input_nodes = []
        self.output_nodes = []

    def draw(self, vars, filename='out.png', prog='dot'):
        for var in vars:
            self.visit(var)

        self.graph.add_subgraph(self.input_nodes, rank='same')

        for rank, nodes in sorted(self.ranks.items(), key=lambda x: x[0]):
            self.graph.add_subgraph(nodes, rank='same')

        self.graph.add_subgraph(self.output_nodes, rank='same')

        self.graph.write('out.dot')
        self.graph.layout(prog=prog)
        self.graph.draw(filename)

    def _set_rank(self, rank, node):
        self.ranks[rank].append(node)

    def _add_output(self, node, src):
        if node._has_output():
            outobj = str(node.output_data)
            label_data = [outobj, str(node.width)]
            if node.point > 0:
                label_data.append(str(node.point))
            label = ':'.join(label_data)
            self.graph.add_node(outobj, label=label, shape='box',
                                color='lightblue', style='filled', peripheries=2)
            self.graph.add_edge(src, outobj)
            self.output_nodes.append(outobj)

    def _add_gap(self, node, mark=''):
        if self.approx:
            return node

        if node.start_stage is None or node.end_stage is None:
            return node

        prev = node
        for i in range(node.end_stage - node.start_stage - 1):
            tmp = self._get_tmp()
            self.graph.add_node(tmp, label=mark, shape='box',
                                color='lightgray', style='filled')
            self.graph.add_edge(prev, tmp)
            self._set_rank(node.start_stage + 2 + i, tmp)
            prev = tmp
        return prev

    def _get_mark(self, obj):
        if obj is None:
            return ''
        if obj.__name__ in vtypes.operator_dict:
            return vtypes.operator_dict[obj.__name__]
        if hasattr(obj, '__call__'):
            return obj.__name__
        return 'custom'

    def _get_tmp(self):
        v = self.tmp_count
        self.tmp_count += 1
        return hash((id(self), v))

    def visit__BinaryOperator(self, node):
        mark = self._get_mark(node.op)
        self.graph.add_node(node, label=mark, shape='circle')

        left = self.visit(node.left)
        right = self.visit(node.right)
        self.graph.add_edge(left, node, label='L')
        self.graph.add_edge(right, node, label='R')

        if node.start_stage is not None:
            self._set_rank(node.start_stage + 1, node)

        prev = self._add_gap(node, mark)
        self._add_output(node, prev)
        return prev

    def visit__UnaryOperator(self, node):
        if self.approx and isinstance(node, stypes._Delay):
            prev = self.visit(node.parent_value)
            self._add_output(node, prev)
            return prev

        mark = ('delay' if isinstance(node, stypes._Delay) else
                'prev' if isinstance(node, stypes._Prev) else
                self._get_mark(node.op))
        shape = 'box' if isinstance(
            node, (stypes._Delay, stypes._Prev)) else 'circle'
        color = 'lightgray' if isinstance(
            node, (stypes._Delay, stypes._Prev)) else 'black'
        style = 'filled' if isinstance(
            node, (stypes._Delay, stypes._Prev)) else None
        self.graph.add_node(node, label=mark, shape=shape,
                            color=color, style=style)

        right = self.visit(node.right)
        self.graph.add_edge(right, node, label='R')

        if node.start_stage is None:
            pass
        elif isinstance(node, stypes._Prev):
            self._set_rank(node.start_stage, node)
        else:
            self._set_rank(node.start_stage + 1, node)

        prev = self._add_gap(node, mark)
        self._add_output(node, prev)
        return prev

    def visit__SpecialOperator(self, node):
        mark = self._get_mark(node.op)
        self.graph.add_node(node, label=mark, shape='ellipse')

        for i, arg in enumerate(node.args):
            a = self.visit(arg)
            self.graph.add_edge(a, node, label=str(i))

        if node.start_stage is not None:
            self._set_rank(node.start_stage + 1, node)

        prev = self._add_gap(node, mark)
        self._add_output(node, prev)
        return prev

    def visit__Accumulator(self, node):
        mark = (' '.join([self._get_mark(op) for op in node.ops])
                if node.label is None else node.label)
        self.graph.add_node(node, label=mark, shape='box', style='rounded')

        right = self.visit(node.right)
        initval = self.visit(node.initval)
        if node.enable is not None:
            enable = self.visit(node.enable)
        if node.reset is not None:
            reset = self.visit(node.reset)
        self.graph.add_edge(right, node, label='R')
        self.graph.add_edge(initval, node, label='initval')
        if node.enable is not None:
            self.graph.add_edge(enable, node, label='enable')
        if node.reset is not None:
            self.graph.add_edge(reset, node, label='reset')

        if node.start_stage is not None:
            self._set_rank(node.start_stage + 1, node)

        prev = self._add_gap(node, mark)
        self._add_output(node, prev)
        return prev

    def visit__ParameterVariable(self, node):
        inobj = str(node.input_data)
        label_data = [inobj, str(node.width)]
        if node.point > 0:
            label_data.append(str(node.point))
        label = ':'.join(label_data)

        self.graph.add_node(node, label=label, shape='',
                            color='lightblue', style='rounded,filled', peripheries=2)

        self.input_nodes.append(node)
        self._add_output(node, node)
        return node

    def visit__Variable(self, node):
        if isinstance(node.input_data, stypes._Numeric):
            input_data = self.visit(node.input_data)
            return input_data

        inobj = str(node.input_data)
        label_data = [inobj, str(node.width)]
        if node.point > 0:
            label_data.append(str(node.point))
        label = ':'.join(label_data)

        self.graph.add_node(node, label=label, shape='box',
                            color='lightblue', style='filled', peripheries=2)

        self.input_nodes.append(node)
        self._add_output(node, node)
        return node

    def visit__Constant(self, node):
        if isinstance(node, stypes.FixedPoint):
            value = "%f:%d" % (
                ((1.0 * node.value) / (2.0 ** node.point)), node.point)
        elif isinstance(node, stypes.Float):
            value = "%f" % node.value
        else:
            value = str(node.value)

        self.graph.add_node(node, label=value, shape='',
                            color='lightblue', style='filled')

        self._add_output(node, node)
        return node



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import collections
import tempfile

import veriloggen.core.vtypes as vtypes
import veriloggen.core.module as module
import veriloggen.core.function as function
import veriloggen.core.task as task

import pyverilog.vparser.ast as vast
from pyverilog.vparser.parser import VerilogCodeParser
from pyverilog.dataflow.modulevisitor import ModuleVisitor
from pyverilog.ast_code_generator.codegen import ASTCodeGenerator


#-------------------------------------------------------------------------
# User interfaces to read Verilog source code
#-------------------------------------------------------------------------
def read_verilog_stubmodule(*filelist, **opt):
    module_dict = to_module_dict(*filelist, **opt)
    codegen = ASTCodeGenerator()
    stubs = collections.OrderedDict()
    for name, m in module_dict.items():
        description = vast.Description((m,))
        source = vast.Source('', description)
        code = codegen.visit(source)
        stubs[name] = module.StubModule(name, code=code)
    return stubs


def read_verilog_module(*filelist, **opt):
    module_dict = to_module_dict(*filelist, **opt)
    visitor = VerilogReadVisitor(module_dict)
    for name, m in module_dict.items():
        visitor.visit(m)
    modules = visitor.converted_modules
    return modules


def read_verilog_module_str(code, encode='utf-8'):
    tmp = tempfile.NamedTemporaryFile()
    tmp.write(code.encode(encode))
    tmp.read()
    filename = tmp.name
    ret = read_verilog_module(filename)
    tmp.close()
    return ret


def read_verilog_stubmodule_str(code, encode='utf-8'):
    tmp = tempfile.NamedTemporaryFile()
    tmp.write(code.encode(encode))
    tmp.read()
    filename = tmp.name
    ret = read_verilog_stubmodule(filename)
    tmp.close()
    return ret


#-------------------------------------------------------------------------
def to_module_dict(*filelist, **opt):
    ast = to_ast(*filelist, **opt)

    module_visitor = ModuleVisitor()
    module_visitor.visit(ast)
    module_names = module_visitor.get_modulenames()
    moduleinfotable = module_visitor.get_moduleinfotable()
    moduleinfo = moduleinfotable.getDefinitions()
    module_dict = collections.OrderedDict(
        [(n, d.definition) for n, d in moduleinfo.items()])

    return module_dict


#-------------------------------------------------------------------------
def to_ast(*filelist, **opt):
    include = opt['include'] if 'include' in opt else ()
    define = opt['define'] if 'define' in opt else ()
    if not isinstance(include, tuple) and not isinstance(include, list):
        raise TypeError('"include" option of read_verilog must be tuple or list, not %s' %
                        type(include))
    if not isinstance(include, tuple) and not isinstance(include, list):
        raise TypeError('"include" option of read_verilog must be tuple or list, not %s' %
                        type(include))

    code_parser = VerilogCodeParser(filelist,
                                    preprocess_include=include,
                                    preprocess_define=define)
    ast = code_parser.parse()

    return ast


#-------------------------------------------------------------------------
def to_tuple(s):
    if not isinstance(s, (list, tuple)):
        return tuple([s])
    return s


#-------------------------------------------------------------------------
class ReadOnlyModule(object):

    def __init__(self, m):
        self.m = m

    def __getattr__(self, attr):
        return getattr(self.m, attr)


#-------------------------------------------------------------------------
class VerilogReadVisitor(object):

    def __init__(self, ast_module_dict, converted_modules=None):
        self.ast_module_dict = ast_module_dict
        self.converted_modules = (collections.OrderedDict()
                                  if converted_modules is None
                                  else converted_modules)
        self.m = None
        self.module_stack = []

    def get_module(self, name):
        if name in self.converted_modules:
            return self.converted_modules[name]
        if name not in self.ast_module_dict:
            return module.StubModule(name)
        visitor = VerilogReadVisitor(
            self.ast_module_dict, self.converted_modules)
        mod = visitor.visit(self.ast_module_dict[name])
        self.converted_modules[name] = mod
        self.converted_modules.update(visitor.converted_modules)
        return mod

    def push_module(self, m):
        self.module_stack.append(self.m)
        self.m = m

    def push_read_only_module(self):
        self.push_module(ReadOnlyModule(self.m))

    def pop_module(self):
        self.m = self.module_stack.pop()

    def add_object(self, obj):
        if isinstance(self.m, module.Module):
            self.m.add_object(obj)
            if isinstance(obj, vtypes._Variable):
                obj.module = self.m

    def generic_visit(self, node):
        for c in node.children():
            self.visit(c)
        #raise TypeError("Unsupported object '%s'" % str(type(node)))

    def visit(self, node):
        method = 'visit_' + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def visit_ModuleDef(self, node):
        # check module cache
        if node.name in self.converted_modules:
            return self.converted_modules[node.name]

        # create new Verilog module
        m = module.Module(node.name)
        self.push_module(m)
        self.generic_visit(node)
        self.pop_module()

        self.converted_modules[node.name] = m
        return m

    def visit_Paramlist(self, node):
        params = []
        for param in node.params:
            p = self.visit(param)
            params.append(p)
        return params

    def visit_Portlist(self, node):
        ports = []
        for port in node.ports:
            p = self.visit(port)
            ports.append(p)
        return ports

    def visit_Port(self, node):
        if node.type is None:
            name = node.name
            p = vtypes.AnyType(name=name)
            self.add_object(p)
            return p

        name = node.name
        width = self.visit(node.width) if node.width is not None else None
        _type = getattr(vtypes, node.type, None)
        if _type is None:
            raise TypeError("No such port type '%s'" % node.type)
        p = _type(name, width)
        self.add_object(p)
        return p

    def visit_Width(self, node):
        msb = self.visit(node.msb)
        width = msb + 1
        return width

    def visit_Length(self, node):
        lsb = self.visit(node.lsb)
        length = lsb + 1
        return length

    def visit_Identifier(self, node):
        if node.scope is not None:
            labels = self.visit(node.scope)
            labels.append(node.name)
            return vtypes.Scope(*labels)
        if not isinstance(self.m, (module.Module, ReadOnlyModule)):
            return vtypes.AnyType(name=node.name)
        ret = self.m.find_identifier(node.name)
        if ret is None:
            return vtypes.AnyType(name=node.name)
        if ret.name in self.m.variable:
            return self.m.variable[ret.name]
        return ret

    def visit_IntConst(self, node):
        return vtypes.Int(node.value)

    def visit_FloatConst(self, node):
        return vtypes.Float(node.value)

    def visit_StringConst(self, node):
        return vtypes.Str(node.value)

    def visit_Input(self, node):
        name = node.name
        width = self.visit(node.width) if node.width is not None else None
        signed = node.signed
        obj = vtypes.Input(width, signed=signed, name=name)
        if node.width is not None:
            obj._set_raw_width(self.visit(node.width.msb),
                               self.visit(node.width.lsb))
        self.add_object(obj)
        return obj

    def visit_Output(self, node):
        name = node.name
        width = self.visit(node.width) if node.width is not None else None
        signed = node.signed
        obj = vtypes.Output(width, signed=signed, name=name)
        if node.width is not None:
            obj._set_raw_width(self.visit(node.width.msb),
                               self.visit(node.width.lsb))
        self.add_object(obj)
        return obj

    def visit_Inout(self, node):
        name = node.name
        width = self.visit(node.width) if node.width is not None else None
        signed = node.signed
        obj = vtypes.Inout(width, signed=signed, name=name)
        if node.width is not None:
            obj._set_raw_width(self.visit(node.width.msb),
                               self.visit(node.width.lsb))
        self.add_object(obj)
        return obj

    def visit_Tri(self, node):
        name = node.name
        width = self.visit(node.width) if node.width is not None else None
        signed = node.signed
        obj = vtypes.Tri(width, signed=signed, name=name)
        if node.width is not None:
            obj._set_raw_width(self.visit(node.width.msb),
                               self.visit(node.width.lsb))
        self.add_object(obj)
        return obj

    def visit_Wire(self, node):
        name = node.name
        width = self.visit(node.width) if node.width is not None else None
        signed = node.signed
        obj = vtypes.Wire(width, signed=signed, name=name)
        if node.width is not None:
            obj._set_raw_width(self.visit(node.width.msb),
                               self.visit(node.width.lsb))
        self.add_object(obj)
        return obj

    def visit_Reg(self, node):
        name = node.name
        width = self.visit(node.width) if node.width is not None else None
        signed = node.signed
        obj = vtypes.Reg(width, signed=signed, name=name)
        if node.width is not None:
            obj._set_raw_width(self.visit(node.width.msb),
                               self.visit(node.width.lsb))
        self.add_object(obj)
        return obj

    def visit_WireArray(self, node):
        name = node.name
        width = self.visit(node.width) if node.width is not None else None
        length = self.visit(node.length)
        signed = node.signed
        obj = vtypes.Wire(width, length=length, signed=signed, name=name)
        if node.width is not None:
            obj._set_raw_width(self.visit(node.width.msb),
                               self.visit(node.width.lsb))
        obj._set_raw_length(self.visit(node.length.msb),
                            self.visit(node.length.lsb))
        self.add_object(obj)
        return obj

    def visit_RegArray(self, node):
        name = node.name
        width = self.visit(node.width) if node.width is not None else None
        length = self.visit(node.length)
        signed = node.signed
        obj = vtypes.Reg(width, length=length, signed=signed, name=name)
        if node.width is not None:
            obj._set_raw_width(self.visit(node.width.msb),
                               self.visit(node.width.lsb))
        obj._set_raw_length(self.visit(node.length.msb),
                            self.visit(node.length.lsb))
        self.add_object(obj)
        return obj

    def visit_Integer(self, node):
        name = node.name
        width = self.visit(node.width) if node.width is not None else None
        signed = node.signed
        obj = vtypes.Integer(width, signed=signed, name=name)
        if node.width is not None:
            obj._set_raw_width(self.visit(node.width.msb),
                               self.visit(node.width.lsb))
        self.add_object(obj)
        return obj

    def visit_Real(self, node):
        name = node.name
        width = self.visit(node.width) if node.width is not None else None
        obj = vtypes.Real(width, name=name)
        if node.width is not None:
            obj._set_raw_width(self.visit(node.width.msb),
                               self.visit(node.width.lsb))
        self.add_object(obj)
        return obj

    def visit_Genvar(self, node):
        name = node.name
        width = self.visit(node.width) if node.width is not None else None
        obj = vtypes.Genvar(width, name=name)
        if node.width is not None:
            obj._set_raw_width(self.visit(node.width.msb),
                               self.visit(node.width.lsb))
        self.add_object(obj)
        return obj

    def visit_Ioport(self, node):
        first = self.visit(node.first)
        second = self.visit(node.second) if node.second is not None else None
        return (first, second)

    def visit_Parameter(self, node):
        name = node.name
        value = self.visit(node.value)
        width = self.visit(node.width) if node.width is not None else None
        signed = node.signed
        param = vtypes.Parameter(value, width, signed, name=name)
        if node.width is not None:
            param._set_raw_width(self.visit(node.width.msb),
                                 self.visit(node.width.lsb))
        self.add_object(param)
        return param

    def visit_Localparam(self, node):
        name = node.name
        value = self.visit(node.value)
        width = self.visit(node.width) if node.width is not None else None
        signed = node.signed
        param = vtypes.Localparam(value, width, signed, name=name)
        if node.width is not None:
            param._set_raw_width(self.visit(node.width.msb),
                                 self.visit(node.width.lsb))
        self.add_object(param)
        return param

    def visit_Supply(self, node):
        name = node.name
        value = self.visit(node.value)
        width = self.visit(node.width) if node.width is not None else None
        signed = node.signed
        param = vtypes.Supply(value, width, signed, name=name)
        if node.width is not None:
            param._set_raw_width(self.visit(node.width.msb),
                                 self.visit(node.width.lsb))
        self.add_object(param)
        return param

    def visit_Decl(self, node):
        decl = [self.visit(d) for d in node.list]
        return decl

    def visit_Concat(self, node):
        vars = [self.visit(var) for var in node.list]
        return vtypes.Cat(*vars)

    def visit_LConcat(self, node):
        vars = [self.visit(var) for var in node.list]
        return vtypes.Cat(*vars)

    def visit_Repeat(self, node):
        var = self.visit(node.value)
        times = self.visit(node.times)
        return vtypes.Repeat(var, times)

    def visit_Partselect(self, node):
        var = self.visit(node.var)
        msb = self.visit(node.msb)
        lsb = self.visit(node.lsb)
        return vtypes.Slice(var, msb, lsb)

    def visit_Pointer(self, node):
        var = self.visit(node.var)
        pos = self.visit(node.ptr)
        return vtypes.Pointer(var, pos)

    def visit_Lvalue(self, node):
        return self.visit(node.var)

    def visit_Rvalue(self, node):
        return self.visit(node.var)

    def visit_Uplus(self, node):
        return vtypes.Uplus(self.visit(node.right))

    def visit_Uminus(self, node):
        return vtypes.Uminus(self.visit(node.right))

    def visit_Ulnot(self, node):
        return vtypes.Ulnot(self.visit(node.right))

    def visit_Unot(self, node):
        return vtypes.Unot(self.visit(node.right))

    def visit_Uand(self, node):
        return vtypes.Uand(self.visit(node.right))

    def visit_Unand(self, node):
        return vtypes.Unand(self.visit(node.right))

    def visit_Uor(self, node):
        return vtypes.Uor(self.visit(node.right))

    def visit_Unor(self, node):
        return vtypes.Unor(self.visit(node.right))

    def visit_Uxor(self, node):
        return vtypes.Uxor(self.visit(node.right))

    def visit_Uxnor(self, node):
        return vtypes.Uxnor(self.visit(node.right))

    def visit_Power(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Power(left, right)

    def visit_Times(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Times(left, right)

    def visit_Divide(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Divide(left, right)

    def visit_Mod(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Mod(left, right)

    def visit_Plus(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Plus(left, right)

    def visit_Minus(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Minus(left, right)

    def visit_Sll(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Sll(left, right)

    def visit_Srl(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Srl(left, right)

    def visit_Sra(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Sra(left, right)

    def visit_LessThan(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.LessThan(left, right)

    def visit_GreaterThan(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.GreaterThan(left, right)

    def visit_LessEq(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.LessEq(left, right)

    def visit_GreaterEq(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.GreaterEq(left, right)

    def visit_Eq(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Eq(left, right)

    def visit_NotEq(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.NotEq(left, right)

    def visit_Eql(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Eql(left, right)

    def visit_NotEql(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.NotEql(left, right)

    def visit_And(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.And(left, right)

    def visit_Xor(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Xor(left, right)

    def visit_Xnor(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Xnor(left, right)

    def visit_Or(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Or(left, right)

    def visit_Land(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Land(left, right)

    def visit_Lor(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vtypes.Lor(left, right)

    def visit_Cond(self, node):
        condition = self.visit(node.cond)
        true_value = self.visit(node.true_value)
        false_value = self.visit(node.false_value)
        return vtypes.Cond(condition, true_value, false_value)

    def visit_Assign(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        ldelay = self.visit(
            node.ldelay.value) if node.ldelay is not None else None
        rdelay = self.visit(
            node.rdelay.value) if node.rdelay is not None else None
        subst = vtypes.Subst(left, right, ldelay=ldelay, rdelay=rdelay)
        assign = vtypes.Assign(subst)
        self.add_object(assign)
        return assign

    def visit_Always(self, node):
        # to avoid to call self.add_object() for the current Module
        self.push_read_only_module()

        sensitivity = self.visit(node.sens_list)
        statement = to_tuple(self.visit(node.statement))
        always = vtypes.Always(*sensitivity)
        always = always(*statement)

        # to restore the current Module
        self.pop_module()

        self.add_object(always)
        return always

    def visit_SensList(self, node):
        return [self.visit(s) for s in node.list]

    def visit_Sens(self, node):
        if node.type == 'posedge':
            sig = self.visit(node.sig)
            return vtypes.Posedge(sig)
        if node.type == 'negedge':
            sig = self.visit(node.sig)
            return vtypes.Negedge(sig)
        if node.type == 'all':
            return vtypes.SensitiveAll()
        if node.type == 'level':
            sig = self.visit(node.sig)
            return sig
        raise TypeError("Unsupported sensitivity type '%s'" % node.type)

    def visit_BlockingSubstitution(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        ldelay = self.visit(
            node.ldelay.delay) if node.ldelay is not None else None
        rdelay = self.visit(
            node.rdelay.delay) if node.rdelay is not None else None
        return vtypes.Subst(left, right, blk=True, ldelay=ldelay, rdelay=rdelay)

    def visit_NonblockingSubstitution(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        ldelay = self.visit(
            node.ldelay.delay) if node.ldelay is not None else None
        rdelay = self.visit(
            node.rdelay.delay) if node.rdelay is not None else None
        return vtypes.Subst(left, right, blk=False, ldelay=ldelay, rdelay=rdelay)

    def visit_IfStatement(self, node):
        if isinstance(self.m, (module.GenerateFor, module.GenerateIf)):
            return self._visit_GenerateIf(node)
        condition = self.visit(node.cond)
        true_statement = self.visit(node.true_statement)
        false_statement = (self.visit(node.false_statement)
                           if node.false_statement is not None else None)
        true_statement = to_tuple(true_statement)
        false_statement = (to_tuple(false_statement)
                           if false_statement is not None else None)
        _if = vtypes.If(condition)
        _if = _if(*true_statement)
        if false_statement is not None:
            _if = _if(*false_statement)
        return _if

    def visit_ForStatement(self, node):
        if isinstance(self.m, (module.GenerateFor, module.GenerateIf)):
            return self._visit_GenerateFor(node)
        pre = self.visit(node.pre)
        condition = self.visit(node.cond)
        post = self.visit(node.post)
        statement = to_tuple(self.visit(node.statement))
        _for = vtypes.For(pre, condition, post)
        _for = _for(*statement)
        return _for

    def visit_WhileStatement(self, node):
        condition = self.visit(node.cond)
        statement = to_tuple(self.visit(node.statement))
        _while = vtypes.While(condition)
        _while = _while(*statement)
        return _while

    def visit_CaseStatement(self, node):
        comp = self.visit(node.comp)
        statement = tuple([self.visit(case) for case in node.caselist])
        case = vtypes.Case(comp)
        case = case(*statement)
        return case

    def visit_CasexStatement(self, node):
        comp = self.visit(node.comp)
        statement = tuple([self.visit(case) for case in node.caselist])
        case = vtypes.Casex(comp)
        case = case(*statement)
        return case

    def visit_Case(self, node):
        condition = tuple([self.visit(c)
                           for c in node.cond]) if node.cond else [None]
        statement = to_tuple(self.visit(node.statement))
        when = vtypes.When(*condition)
        when = when(*statement)
        return when

    def visit_Block(self, node):
        statements = [self.visit(statement) for statement in node.statements]
        return statements

    def visit_Initial(self, node):
        statement = to_tuple(self.visit(node.statement))
        initial = vtypes.Initial(*statement)
        self.add_object(initial)
        return initial

    def visit_EventStatement(self, node):
        sensitivity = self.visit(node.sens_list)
        event = vtypes.Event(*sensitivity)
        return event

    def visit_WaitStatement(self, node):
        condition = self.visit(node.cond)
        statement = to_tuple(self.visit(node.statement)
                             ) if node.statement else None
        wait = vtypes.Wait(condition)
        if statement:
            wait = wait(*statement)
        return wait

    def visit_ForeverStatement(self, node):
        statement = to_tuple(self.visit(node.statement))
        forever = vtypes.Forever(*statement)
        return forever

    def visit_DelayStatement(self, node):
        value = self.visit(node.delay)
        delay = vtypes.Delay(value)
        return delay

    def visit_InstanceList(self, node):
        return [self.visit(instance) for instance in node.instances]

    def visit_Instance(self, node):
        m = self.get_module(node.module)
        instname = node.name
        params = [self.visit(param) for param in node.parameterlist]
        ports = [self.visit(port) for port in node.portlist]
        if node.array is not None:
            raise ValueError("Instance array is not currently supported.")
        instance = module.Instance(m, instname, params, ports)
        self.add_object(instance)
        return instance

    def visit_ParamArg(self, node):
        paramname = node.paramname
        argname = self.visit(node.argname)
        return (paramname, argname)

    def visit_PortArg(self, node):
        portname = node.portname
        argname = self.visit(
            node.argname) if node.argname is not None else None
        return (portname, argname)

    def visit_Function(self, node):
        # to avoid to call self.add_object() for the current Module
        self.push_read_only_module()

        name = node.name
        width = self.visit(
            node.retwidth) if node.retwidth is not None else None
        func = function.Function(name, width)
        if node.retwidth is not None:
            func._set_raw_width(self.visit(node.retwidth.msb),
                                self.visit(node.retwidth.lsb))
        statement = [self.visit(s) for s in node.statement]
        body = []

        for s in statement:
            if isinstance(s, (tuple, list)):  # from the visitor result of decl
                for d in s:
                    if isinstance(d, vtypes.Input):
                        t = func.Input(d.name, d.width,
                                       d.length, d.signed, d.value)
                        if d.width_msb is not None and d.width_lsb is not None:
                            t._set_raw_width(d.width_msb, d.width_lsb)
                    elif isinstance(d, vtypes.Reg):
                        t = func.Reg(d.name, d.width, d.length,
                                     d.signed, d.value)
                        if d.width_msb is not None and d.width_lsb is not None:
                            t._set_raw_width(d.width_msb, d.width_lsb)
                    elif isinstance(d, vtypes.Integer):
                        t = func.Integer(
                            d.name, d.width, d.length, d.signed, d.value)
                        if d.width_msb is not None and d.width_lsb is not None:
                            t._set_raw_width(d.width_msb, d.width_lsb)
                    else:
                        body.append(s)
            else:
                body.append(s)

        func.Body(*body)

        # to restore the current Module
        self.pop_module()

        self.add_object(func)
        return func

    def visit_FunctionCall(self, node):
        name = self.visit(node.name)
        args = tuple([self.visit(arg) for arg in node.args])
        call = function.FunctionCall(name, *args)
        return call

    def visit_Task(self, node):
        # to avoid to call self.add_object() for the current Module
        self.push_read_only_module()

        name = node.name
        _task = task.Task(name)
        statement = [self.visit(s) for s in node.statement]
        body = []

        for s in statement:
            if isinstance(s, (tuple, list)):  # from the visitor result of decl
                for d in s:
                    if isinstance(d, vtypes.Input):
                        t = _task.Input(d.name, d.width,
                                        d.length, d.signed, d.value)
                        if d.width_msb is not None and d.width_lsb is not None:
                            t._set_raw_width(d.width_msb, d.width_lsb)
                    elif isinstance(d, vtypes.Reg):
                        t = _task.Reg(d.name, d.width, d.length,
                                      d.signed, d.value)
                        if d.width_msb is not None and d.width_lsb is not None:
                            t._set_raw_width(d.width_msb, d.width_lsb)
                    elif isinstance(d, vtypes.Integer):
                        t = _task.Integer(
                            d.name, d.width, d.length, d.signed, d.value)
                        if d.width_msb is not None and d.width_lsb is not None:
                            t._set_raw_width(d.width_msb, d.width_lsb)
                    else:
                        body.append(s)
            else:
                body.append(s)

        _task.Body(*body)

        # to restore the current Module
        self.pop_module()

        self.add_object(_task)
        return _task

    def visit_TaskCall(self, node):
        name = self.visit(node.name)
        args = tuple([self.visit(arg) for arg in node.args])
        call = task.TaskCall(name, args)
        return call

    def _visit_GenerateFor(self, item):
        pre = self.visit(item.pre)
        cond = self.visit(item.cond)
        post = self.visit(item.post)
        scope = (item.statement.scope
                 if isinstance(item.statement, vast.Block)
                 else None)
        _for = module.GenerateFor(self.m, pre, cond, post, scope)
        ret = _for
        self.add_object(_for)
        self.push_module(_for)
        statement = self.visit(item.statement)
        self.pop_module()
        return ret

    def _visit_GenerateIf(self, item):
        cond = self.visit(item.cond)
        true_scope = (item.true_statement.scope
                      if isinstance(item.true_statement, vast.Block)
                      else None)
        false_scope = (item.false_statement.scope
                       if isinstance(item.false_statement, vast.Block)
                       else None)
        _if_true = module.GenerateIf(self.m, cond, true_scope)
        ret = _if_true
        self.add_object(_if_true)
        self.push_module(_if_true)
        statement = self.visit(item.true_statement)
        self.pop_module()
        _if_false = _if_true.Else(false_scope)
        self.push_module(_if_false)
        statement = (self.visit(item.false_statement)
                     if item.false_statement is not None else None)
        self.pop_module()
        return ret

    def visit_GenerateStatement(self, node):
        ret = []
        for item in node.items:
            if isinstance(item, vast.ForStatement):
                ret.append(self._visit_GenerateFor(item))
            elif isinstance(item, vast.IfStatement):
                ret.append(self._visit_GenerateIf(item))
            else:
                raise TypeError(
                    "Only generate-for and generate-if statements are supported.")
        return ret

    def visit_SystemCall(self, node):
        cmd = node.syscall
        args = tuple([self.visit(arg) for arg in node.args])
        systask = vtypes.SystemTask(cmd, *args)
        return systask

    def visit_IdentifierScopeLabel(self, node):
        if node.loop is None:
            return node.name
        index = self.visit(node.loop)
        return vtypes.ScopeIndex(node.name, index)

    def visit_IdentifierScope(self, node):
        args = [self.visit(label) for label in node.labellist]
        return args

    def visit_Pragma(self, node):
        raise TypeError("Pragma is not currently supported.")

    def visit_PragmaEntry(self, node):
        raise TypeError("Pragma is not currently supported.")

    def visit_Disable(self, node):
        raise TypeError("Disable is not currently supported.")

    def visit_ParallelBlock(self, node):
        raise TypeError("Fork/Join is not currently supported.")

    def visit_SingleStatement(self, node):
        statement = self.visit(node.statement)
        if isinstance(statement, vtypes.Delay):
            return statement
        return vtypes.SingleStatement(statement)



#-------------------------------------------------------------------------
# Veriloggen: A library for constructing a Verilog HDL source code in Python
#
# Copyright (C) 2015, Shinya Takamaeda-Yamazaki
# License: Apache 2.0
#-------------------------------------------------------------------------
from __future__ import absolute_import
from __future__ import print_function

# expand the recursive call limit for python 3.6 and later
import sys
sys.setrecursionlimit(1000 * 10)

# Verilog HDL Core
from .core.vtypes import *
from .core.module import Module, StubModule, Instance, GenerateFor, GenerateIf, connect_same_name
from .core.function import Function, FunctionCall
from .core.task import Task, TaskCall
from .core.submodule import Submodule

# Code Generator
from .verilog import from_verilog

# Verilog Simulator
from .simulation import simulation

# Abstract Extension
from .seq.seq import Seq, TmpSeq, make_condition
from .fsm.fsm import FSM, TmpFSM
from .pipeline.pipeline import Pipeline

# Extension reset
from .seq import reset as seq_reset
from .fsm import reset as fsm_reset
from .dataflow import reset as dataflow_reset
from .stream import reset as stream_reset
from .thread import reset as thread_reset
from .types import reset as types_reset


def reset():
    seq_reset()
    fsm_reset()
    dataflow_reset()
    stream_reset()
    thread_reset()
    types_reset()



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import collections
import re

import veriloggen.core.vtypes as vtypes
import veriloggen.core.module as module

import pyverilog.vparser.ast as vast
from pyverilog.ast_code_generator.codegen import ASTCodeGenerator


#-------------------------------------------------------------------------
def write_verilog(node, filename=None, for_verilator=False):
    visitor = VerilogModuleVisitor(for_verilator)
    modules = tuple(node.get_modules().values())

    module_ast_list = [visitor.visit(mod) for mod in modules
                       if not isinstance(mod, module.StubModule)]
    description = vast.Description(module_ast_list)
    source = vast.Source(filename, description)

    codegen = ASTCodeGenerator()
    main = codegen.visit(source)

    stub = [mod.get_code()
            for mod in modules if isinstance(mod, module.StubModule)]

    code = ''.join([main] + stub)

    if filename:
        with open(filename, 'w') as f:
            f.write(code)
    return code


#-------------------------------------------------------------------------
class VerilogCommonVisitor(object):

    def __init__(self, for_verilator=False, in_initial=False):
        self.for_verilator = for_verilator
        self.in_initial = in_initial

    def generic_visit(self, node):
        raise TypeError("Type %s is not supported." % str(type(node)))

    def visit(self, node):
        name = node.__class__.__name__
        if hasattr(node, 'ast_name') and node.ast_name is not None:
            name = node.ast_name
        visitor = getattr(self, 'visit_' + name, self.generic_visit)
        return visitor(node)

    #-------------------------------------------------------------------------
    # First class object wrapper
    def visit_Int(self, node):
        value_list = []
        value = node.value

        if isinstance(node.value, str) and node.value.find('-') >= 0:
            pos = node.value.find('-') + 1
            value = node.value[pos:]
            value_list.append('-')

        if isinstance(node.value, int) and node.value < 0:
            value = abs(node.value)
            value_list.append('-')

        if node.width:
            value_list.append(str(node.width))

        if node.base is None:
            if node.signed:
                value_list.append("'sd")
            elif node.width:
                value_list.append("'d")
            if isinstance(value, str):
                value_list.append(value)
            else:
                value_list.append(str(value))
        elif node.base == 2:
            if node.signed:
                value_list.append("'sb")
            else:
                value_list.append("'b")
            if isinstance(value, str):
                value_list.append(value)
            else:
                value_list.append(bin(value).replace('0b', ''))
        elif node.base == 8:
            if node.signed:
                value_list.append("'so")
            else:
                value_list.append("'o")
            if isinstance(value, str):
                value_list.append(value)
            else:
                value_list.append(oct(value).replace('0o', ''))
        elif node.base == 10:
            if node.signed:
                value_list.append("'sd")
            else:
                value_list.append("'d")
            if isinstance(value, str):
                value_list.append(value)
            else:
                value_list.append(str(value))
        elif node.base == 16:
            if node.signed:
                value_list.append("'sh")
            else:
                value_list.append("'h")
            if isinstance(value, str):
                value_list.append(value)
            else:
                value_list.append(hex(value).replace('0x', ''))
        else:
            raise ValueError("Int.base must be 2, 8, 10, or 16")

        return vast.IntConst(''.join(value_list))

    def visit_Float(self, node):
        return vast.FloatConst(str(node.value))

    def visit_Str(self, node):
        return vast.StringConst(node.value)

    #-------------------------------------------------------------------------
    def visit_bool(self, node):
        if node:
            return vast.IntConst('1')
        return vast.IntConst('0')

    def visit_int(self, node):
        return vast.IntConst(str(node))

    def visit_str(self, node):
        return vast.StringConst(node)

    def visit_float(self, node):
        return vast.FloatConst(str(node))

    def visit_list(self, node):
        return self._optimize_block(vast.Block(tuple([self.visit(n) for n in node])))

    def visit_tuple(self, node):
        return self.visit_list(node)

    def _optimize_block(self, node):
        if not isinstance(node, vast.Block):
            return node
        ret = []
        for n in node.statements:
            if isinstance(n, vast.Block) and n.scope is None:
                ret.extend(n.statements)
            else:
                ret.append(n)
        return vast.Block(tuple(ret))

    #-------------------------------------------------------------------------
    def visit_Parameter(self, node):
        name = node.name
        return vast.Identifier(name)

    def visit_Localparam(self, node):
        name = node.name
        return vast.Identifier(name)

    #-------------------------------------------------------------------------
    def visit_Input(self, node):
        name = node.name
        return vast.Identifier(name)

    def visit_Output(self, node):
        name = node.name
        return vast.Identifier(name)

    def visit_Inout(self, node):
        name = node.name
        return vast.Identifier(name)

    def visit_Reg(self, node):
        name = node.name
        return vast.Identifier(name)

    def visit_Wire(self, node):
        name = node.name
        return vast.Identifier(name)

    def visit_Integer(self, node):
        name = node.name
        return vast.Identifier(name)

    def visit_Real(self, node):
        name = node.name
        return vast.Identifier(name)

    def visit_Genvar(self, node):
        name = node.name
        return vast.Identifier(name)

    def visit_AnyType(self, node):
        name = node.name
        return vast.Identifier(name)

    #-------------------------------------------------------------------------
    def visit__SkipUnaryOperator(self, node):
        return self.visit(node.right)

    #-------------------------------------------------------------------------
    def visit_Power(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Power(left, right)

    def visit_Times(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Times(left, right)

    def visit_Divide(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Divide(left, right)

    def visit_Mod(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Mod(left, right)

    def visit_Plus(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Plus(left, right)

    def visit_Minus(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Minus(left, right)

    def visit_Sll(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Sll(left, right)

    def visit_Srl(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Srl(left, right)

    def visit_Sra(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Sra(left, right)

    def visit_LessThan(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.LessThan(left, right)

    def visit_GreaterThan(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.GreaterThan(left, right)

    def visit_LessEq(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.LessEq(left, right)

    def visit_GreaterEq(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.GreaterEq(left, right)

    def visit_Eq(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Eq(left, right)

    def visit_NotEq(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.NotEq(left, right)

    def visit_Eql(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Eql(left, right)

    def visit_NotEql(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.NotEql(left, right)

    def visit_And(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.And(left, right)

    def visit_Xor(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Xor(left, right)

    def visit_Xnor(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Xnor(left, right)

    def visit_Or(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Or(left, right)

    def visit_Land(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Land(left, right)

    def visit_Lor(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return vast.Lor(left, right)

    #-------------------------------------------------------------------------
    def visit_Uplus(self, node):
        right = self.visit(node.right)
        return vast.Uplus(right)

    def visit_Uminus(self, node):
        right = self.visit(node.right)
        return vast.Uminus(right)

    def visit_Ulnot(self, node):
        right = self.visit(node.right)
        return vast.Ulnot(right)

    def visit_Unot(self, node):
        right = self.visit(node.right)
        return vast.Unot(right)

    def visit_Uand(self, node):
        right = self.visit(node.right)
        return vast.Uand(right)

    def visit_Unand(self, node):
        right = self.visit(node.right)
        return vast.Unand(right)

    def visit_Uor(self, node):
        right = self.visit(node.right)
        return vast.Uor(right)

    def visit_Unor(self, node):
        right = self.visit(node.right)
        return vast.Unor(right)

    def visit_Uxor(self, node):
        right = self.visit(node.right)
        return vast.Uxor(right)

    def visit_Uxnor(self, node):
        right = self.visit(node.right)
        return vast.Uxnor(right)

   #---------------------------------------------------------------------------
    def visit_Pointer(self, node):
        var = self.visit(node.var)
        pos = self.visit(node.pos)
        return vast.Pointer(var, pos)

    def visit_Slice(self, node):
        var = self.visit(node.var)
        msb = self.visit(node.msb)
        lsb = self.visit(node.lsb)
        return vast.Partselect(var, msb, lsb)

    def visit_Cat(self, node):
        vars = tuple([self.visit(var) for var in node.vars])
        return vast.Concat(vars)

    def visit_Repeat(self, node):
        var = (self.visit(node.var) if isinstance(node.var, vtypes.Cat) else
               self.visit(vtypes.Cat(node.var)))
        times = self.visit(node.times)
        return vast.Repeat(var, times)

    #-------------------------------------------------------------------------
    def visit_Cond(self, node):
        cond = self.visit(node.condition)
        true_value = self.visit(node.true_value)
        false_value = self.visit(node.false_value)
        return vast.Cond(cond, true_value, false_value)

    #-------------------------------------------------------------------------
    def visit_If(self, node):
        cond = self.visit(node.condition)
        true_statement = self.visit(node.true_statement)
        false_statement = (self.visit(node.false_statement)
                           if node.false_statement is not None else None)
        # remove a redundant begin-end, if the true_statement is not for reset
        cond_str = str(cond).lower()
        if ((not cond_str.count('rst') and not cond_str.count('reset')) and
            isinstance(false_statement, vast.Block) and
            len(false_statement.statements) == 1 and
                isinstance(false_statement.statements[0], vast.IfStatement)):
            false_statement = false_statement.statements[0]
        return vast.IfStatement(cond, true_statement, false_statement)

    #-------------------------------------------------------------------------
    def visit_For(self, node):
        for_visitor = VerilogBlockingVisitor(self.for_verilator, self.in_initial)
        pre = for_visitor.visit(node.pre)
        cond = self.visit(node.condition)
        post = for_visitor.visit(node.post)
        statement = self.visit(node.statement)
        return vast.ForStatement(pre, cond, post, statement)

    #-------------------------------------------------------------------------
    def visit_While(self, node):
        cond = self.visit(node.condition)
        statement = self.visit(node.statement)
        return vast.WhileStatement(cond, statement)

    #-------------------------------------------------------------------------
    def visit_Case(self, node):
        comp = self.visit(node.comp)
        statement = tuple([self.visit(s) for s in node.statement])
        return vast.CaseStatement(comp, statement)

    def visit_Casex(self, node):
        comp = self.visit(node.comp)
        statement = tuple([self.visit(s) for s in node.statement])
        return vast.CasexStatement(comp, statement)

    def visit_When(self, node):
        cond = tuple([self.visit(c)
                      for c in node.condition]) if node.condition else None
        statement = self.visit(node.statement)
        return vast.Case(cond, statement)

    #-------------------------------------------------------------------------
    def visit_ScopeIndex(self, node):
        name = node.name
        index = self.visit(node.index)
        return vast.IdentifierScopeLabel(name, index)

    def visit_Scope(self, node):
        scope = []
        for a in node.args:
            if isinstance(a, module.GenerateIf):
                if a.true_scope is None:
                    raise TypeError("GenerateIf statement without scope name"
                                    "can not be used as an scope index.")
                scope.append(vast.IdentifierScopeLabel(a.true_scope))
            elif isinstance(a, module.GenerateIfElse):
                if a.false_scope is None:
                    raise TypeError("GenerateIfElse statement without scope name"
                                    "can not be used as an scope index.")
                scope.append(vast.IdentifierScopeLabel(a.false_scope))
            elif isinstance(a, module.GenerateFor):
                raise TypeError("Scope index must not be GenerateFor. "
                                "Use slice, like 'obj[index]'")
            elif isinstance(a, vtypes.ScopeIndex):
                scope.append(self.visit(a))
            elif isinstance(a, str):
                scope.append(vast.IdentifierScopeLabel(a))
            else:
                _id = self.visit(a)
                if not isinstance(_id, vast.Identifier):
                    raise TypeError("Cannot convert into IdentifierScopeLabel from %s." %
                                    str(type(_id)))
                scope.append(vast.IdentifierScopeLabel(_id.name))

        if len(scope) == 1:
            return vast.Identifier(scope[0].name)
        return vast.Identifier(scope[-1].name, vast.IdentifierScope(tuple(scope[:-1])))

    #-------------------------------------------------------------------------
    def visit_SystemTask(self, node):
        cmd = node.cmd
        if (self.for_verilator and (cmd == 'dumpfile' or cmd == 'dumpvars')):
            return vast.SystemCall('write', (vast.StringConst(''),))
        if (self.for_verilator and self.in_initial and cmd == 'finish'):
            return vast.SystemCall('write', (vast.StringConst(''),))

        args = tuple([self.visit(a) for a in node.args])
        return vast.SystemCall(cmd, args)

    def visit_SingleStatement(self, node):
        statement = self.visit(node.statement)
        return vast.SingleStatement(statement)

    #-------------------------------------------------------------------------
    def visit_Event(self, node):
        sensitivity = vast.SensList(
            tuple([self.visit(n) if isinstance(n, vtypes.Sensitive) else
                   vast.Sens(self.visit(n)) for n in node.sensitivity]))
        return vast.EventStatement(sensitivity)

    def visit_Delay(self, node):
        if self.for_verilator:
            return vast.SingleStatement(vast.SystemCall('write', (vast.StringConst(''),)))
        delay = self.visit(node.value)
        return vast.SingleStatement(vast.DelayStatement(delay))

    #-------------------------------------------------------------------------
    def visit_Function(self, node):
        name = node.name
        return vast.Identifier(name)

    def visit_FunctionCall(self, node):
        func = self.visit(node.func)
        args = tuple([self.visit(a) for a in node.args])
        return vast.FunctionCall(func, args)

    #-------------------------------------------------------------------------
    def visit_Task(self, node):
        name = node.name
        return vast.Identifier(name)

    def visit_TaskCall(self, node):
        name = vast.Identifier(node.name)
        args = tuple([self.visit(a) for a in node.args])
        return vast.TaskCall(name, args)

    #-------------------------------------------------------------------------
    def visit_Instance(self, node):
        return vast.Identifier(node.instname)

    #-------------------------------------------------------------------------
    def visit_EmbeddedCode(self, node):
        return vast.EmbeddedCode(node.code)

    def visit_EmbeddedNumeric(self, node):
        return vast.EmbeddedCode(node.code)


#-------------------------------------------------------------------------
class VerilogModuleVisitor(VerilogCommonVisitor):

    def __init__(self, for_verilator=False):
        VerilogCommonVisitor.__init__(self, for_verilator)
        self.bind_visitor = VerilogBindVisitor(for_verilator)
        self.always_visitor = VerilogAlwaysVisitor(for_verilator)
        self.blocking_visitor = VerilogBlockingVisitor(for_verilator)
        self.module = None

    #-------------------------------------------------------------------------
    def make_width(self, node):
        msb = vast.Minus(self.bind_visitor.visit(node), vast.IntConst('1'))
        lsb = vast.IntConst('0')
        return vast.Width(msb, lsb)

    def make_width_msb_lsb(self, msb, lsb):
        msb = self.bind_visitor.visit(msb)
        lsb = self.bind_visitor.visit(lsb)
        return vast.Width(msb, lsb)

    def make_length(self, node):
        msb = vast.IntConst('0')
        lsb = vast.Minus(self.bind_visitor.visit(node), vast.IntConst('1'))
        return vast.Length(msb, lsb)

    def make_length_msb_lsb(self, msb, lsb):
        msb = self.bind_visitor.visit(msb)
        lsb = self.bind_visitor.visit(lsb)
        return vast.Length(msb, lsb)

    #-------------------------------------------------------------------------
    def visit(self, node):
        if isinstance(node, module.Module) and not isinstance(node, module.Generate):
            return self.visit_Module(node)
        name = node.__class__.__name__
        if hasattr(node, 'ast_name') and node.ast_name is not None:
            name = node.ast_name
        visitor = getattr(self, 'visit_' + name, self.generic_visit)
        return visitor(node)

    #-------------------------------------------------------------------------
    def visit_Module(self, node):
        self.module = node
        name = node.name

        params = ([self.visit(v) for v in node.global_constant.values()])
        params = [i for i in params if i is not None]
        paramlist = vast.Paramlist(tuple(params))

        ports = [self.visit(v) for v in node.io_variable.values()]
        ports = [i for i in ports if i is not None]
        portlist = vast.Portlist(tuple(ports))

        excludes = [vtypes.Input, vtypes.Output,
                    vtypes.Inout, vtypes.Parameter]

        items = [self.visit(i) for i in node.items
                 if not isinstance(i, tuple(excludes))]
        items = [i for i in items if i is not None]

        m = vast.ModuleDef(name, paramlist, portlist, items)

        self.module = None
        return m

    #-------------------------------------------------------------------------
    def visit_Parameter(self, node):
        name = node.name
        value = self.bind_visitor.visit(node.value)
        value = vast.Rvalue(value)
        width = (None if node.width is None else
                 self.make_width_msb_lsb(node.width_msb, node.width_lsb)
                 if node.width_msb is not None and node.width_lsb is not None else
                 self.make_width(node.width))
        signed = node.signed
        return vast.Parameter(name, value, width, signed)

    def visit_Localparam(self, node):
        name = node.name
        value = self.bind_visitor.visit(node.value)
        value = vast.Rvalue(value)
        width = (None if node.width is None else
                 self.make_width_msb_lsb(node.width_msb, node.width_lsb)
                 if node.width_msb is not None and node.width_lsb is not None else
                 self.make_width(node.width))
        signed = node.signed
        return vast.Localparam(name, value, width, signed)

    #-------------------------------------------------------------------------
    def visit_Input(self, node):
        name = node.name
        width = (None if node.width is None else
                 self.make_width_msb_lsb(node.width_msb, node.width_lsb)
                 if node.width_msb is not None and node.width_lsb is not None else
                 self.make_width(node.width))
        signed = node.signed
        return vast.Ioport(vast.Input(name, width, signed))

    def visit_Output(self, node):
        name = node.name
        width = (None if node.width is None else
                 self.make_width_msb_lsb(node.width_msb, node.width_lsb)
                 if node.width_msb is not None and node.width_lsb is not None else
                 self.make_width(node.width))
        signed = node.signed
        first = vast.Output(name, width, signed)
        second = vast.Reg(name, width, signed) if self.module.is_reg(
            name) else None
        return vast.Ioport(first, second)

    def visit_Inout(self, node):
        name = node.name
        width = (None if node.width is None else
                 self.make_width_msb_lsb(node.width_msb, node.width_lsb)
                 if node.width_msb is not None and node.width_lsb is not None else
                 self.make_width(node.width))
        signed = node.signed
        return vast.Ioport(vast.Inout(name, width, signed))

    def visit_Reg(self, node):
        name = node.name
        if self.module.is_output(name):
            return None
        width = (None if node.width is None else
                 self.make_width_msb_lsb(node.width_msb, node.width_lsb)
                 if node.width_msb is not None and node.width_lsb is not None else
                 self.make_width(node.width))
        length = (None if node.length is None else
                  self.make_length_msb_lsb(node.length_msb, node.length_lsb)
                  if node.length_msb is not None and node.length_lsb is not None else
                  self.make_length(node.length))
        signed = node.signed
        if length is not None:
            return vast.RegArray(name, width, length, signed)
        return vast.Reg(name, width, signed)

    def visit_Wire(self, node):
        name = node.name
        width = (None if node.width is None else
                 self.make_width_msb_lsb(node.width_msb, node.width_lsb)
                 if node.width_msb is not None and node.width_lsb is not None else
                 self.make_width(node.width))
        length = (None if node.length is None else
                  self.make_length_msb_lsb(node.length_msb, node.length_lsb)
                  if node.length_msb is not None and node.length_lsb is not None else
                  self.make_length(node.length))
        signed = node.signed
        if length is not None:
            return vast.WireArray(name, width, length, signed)
        return vast.Wire(name, width, signed)

    def visit_Integer(self, node):
        name = node.name
        width = (None if node.width is None else
                 self.make_width_msb_lsb(node.width_msb, node.width_lsb)
                 if node.width_msb is not None and node.width_lsb is not None else
                 self.make_width(node.width))
        signed = node.signed
        return vast.Integer(name, width, signed)

    def visit_Real(self, node):
        name = node.name
        width = (None if node.width is None else
                 self.make_width_msb_lsb(node.width_msb, node.width_lsb)
                 if node.width_msb is not None and node.width_lsb is not None else
                 self.make_width(node.width))
        signed = node.signed
        return vast.Real(name, width, signed)

    def visit_Genvar(self, node):
        name = node.name
        width = (None if node.width is None else
                 self.make_width_msb_lsb(node.width_msb, node.width_lsb)
                 if node.width_msb is not None and node.width_lsb is not None else
                 self.make_width(node.width))
        signed = node.signed
        return vast.Genvar(name, width, signed)

    def visit_AnyType(self, node):
        return None

    #-------------------------------------------------------------------------
    def visit_Posedge(self, node):
        sig = self.bind_visitor.visit(node.name)
        t = 'posedge'
        return vast.Sens(sig, t)

    def visit_Negedge(self, node):
        sig = self.bind_visitor.visit(node.name)
        t = 'negedge'
        return vast.Sens(sig, t)

    def visit_SensitiveAll(self, node):
        sig = None
        t = 'all'
        return vast.Sens(sig, t)

    #-------------------------------------------------------------------------
    def visit_Always(self, node):
        sens = (tuple([self.visit(n) if isinstance(n, vtypes.Sensitive) else
                       vast.Sens(self.always_visitor.visit(n), 'level')
                       for n in node.sensitivity]) if node.sensitivity else
                tuple([vast.Sens(None, 'all')]))
        sensitivity = vast.SensList(sens)
        statement = self._optimize_block(
            vast.Block(tuple([self.always_visitor.visit(n) for n in node.statement])))
        return vast.Always(sensitivity, statement)

    #-------------------------------------------------------------------------
    def visit_Assign(self, node):
        if not isinstance(node.statement, vtypes.Subst):
            raise TypeError("Assign expects Subst object.")
        left = self.bind_visitor.visit(node.statement.left)
        right = self.bind_visitor.visit(node.statement.right)
        lvalue = vast.Lvalue(left)
        rvalue = vast.Rvalue(right)
        return vast.Assign(lvalue, rvalue)

    #-------------------------------------------------------------------------
    def visit_Initial(self, node):
        self.blocking_visitor.in_initial = True
        statement = self._optimize_block(
            vast.Block(tuple([self.blocking_visitor.visit(s) for s in node.statement])))
        self.blocking_visitor.in_initial = False
        return vast.Initial(statement)

    #-------------------------------------------------------------------------
    def visit_Function(self, node):
        name = node.name
        retwidth = (None if node.width is None else
                    self.make_width_msb_lsb(node.width_msb, node.width_lsb)
                    if node.width_msb is not None and node.width_lsb is not None else
                    self.make_width(node.width))
        statement = ([self.visit(v).first for v in node.io_variable.values()] +
                     [self.visit(v) for v in node.variable.values()])
        statement.append(self._optimize_block(
            vast.Block(tuple([self.blocking_visitor.visit(s) for s in node.statement]))))
        return vast.Function(name, retwidth, statement)

    #-------------------------------------------------------------------------
    def visit_Task(self, node):
        name = node.name
        statement = ([self.visit(v).first for v in node.io_variable.values()] +
                     [self.visit(v) for v in node.variable.values()])
        statement.append(self._optimize_block(
            vast.Block(tuple([self.blocking_visitor.visit(s) for s in node.statement]))))
        return vast.Task(name, statement)

    #-------------------------------------------------------------------------
    def visit_Instance(self, node):
        module = node.module.name
        parameterlist = [vast.ParamArg(
            p, self.bind_visitor.visit(a)) for p, a in node.params]
        portlist = [vast.PortArg(p, self.bind_visitor.visit(a))
                    for p, a in node.ports]
        name = node.instname
        instance = vast.Instance(module, name, portlist, parameterlist)
        return vast.InstanceList(module, parameterlist, (instance,))

    #-------------------------------------------------------------------------
    def _visit_Generate(self, node):
        params = ([self.visit(v) for v in node.global_constant.values()])
        params = [i for i in params if i is not None]
        paramlist = [vast.Decl(p) for p in params]
        items = [self.visit(i) for i in node.items
                 if not isinstance(i, (vtypes.Input, vtypes.Output,
                                       vtypes.Inout, vtypes.Parameter))]
        items = [i for i in items if i is not None]
        ret = paramlist
        for i in items:
            if isinstance(i, vast.GenerateStatement):
                ret.extend(i.items)
            else:
                ret.append(i)
        return tuple(ret)

    def visit_GenerateFor(self, node):
        genfor_visitor = VerilogBlockingVisitor(self.for_verilator)
        pre = genfor_visitor.visit(node.pre)
        cond = genfor_visitor.visit(node.cond)
        post = genfor_visitor.visit(node.post)
        items = self._visit_Generate(node)
        block = vast.Block(items, scope=node.scope)
        _for = vast.ForStatement(pre, cond, post, block)
        return vast.GenerateStatement(tuple([_for]))

    def visit_GenerateIf(self, node):
        genif_visitor = VerilogBlockingVisitor(self.for_verilator)
        cond = genif_visitor.visit(node.cond)
        true_items = self._visit_Generate(node)
        true_block = vast.Block(true_items, scope=node.true_scope)
        false_items = self._visit_Generate(node.Else)
        false_block = (vast.Block(false_items, scope=node.Else.false_scope)
                       if false_items else None)
        _if = vast.IfStatement(cond, true_block, false_block)
        return vast.GenerateStatement(tuple([_if]))


#-------------------------------------------------------------------------
class VerilogBindVisitor(VerilogCommonVisitor):
    pass


#-------------------------------------------------------------------------
class VerilogAlwaysVisitor(VerilogCommonVisitor):

    def visit_Subst(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        ldelay = self.visit(node.ldelay) if node.ldelay else None
        rdelay = self.visit(node.rdelay) if node.rdelay else None
        lvalue = vast.Lvalue(left)
        rvalue = vast.Rvalue(right)
        if node.blk:
            return vast.BlockingSubstitution(lvalue, rvalue, ldelay, rdelay)
        return vast.NonblockingSubstitution(lvalue, rvalue, ldelay, rdelay)


#-------------------------------------------------------------------------
class VerilogBlockingVisitor(VerilogCommonVisitor):

    def visit_Subst(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        ldelay = vast.DelayStatement(self.visit(
            node.ldelay)) if node.ldelay else None
        rdelay = vast.DelayStatement(self.visit(
            node.rdelay)) if node.rdelay else None
        lvalue = vast.Lvalue(left)
        rvalue = vast.Rvalue(right)
        return vast.BlockingSubstitution(lvalue, rvalue, ldelay, rdelay)

    def visit_Forever(self, node):
        if self.for_verilator:
            return vast.SingleStatement(vast.SystemCall('write', (vast.StringConst(''),)))
        statement = self._optimize_block(
            vast.Block(tuple([self.visit(s) for s in node.statement])))
        return vast.ForeverStatement(statement)

    def visit_Wait(self, node):
        cond = self.visit(node.condition)
        if node.statement is None:
            return vast.WaitStatement(cond, None)
        statement = self._optimize_block(
            vast.Block(tuple([self.visit(s) for s in node.statement])))
        return vast.WaitStatement(cond, statement)

    def visit_Posedge(self, node):
        sig = self.visit(node.name)
        t = 'posedge'
        return vast.Sens(sig, t)

    def visit_Negedge(self, node):
        sig = self.visit(node.name)
        t = 'negedge'
        return vast.Sens(sig, t)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

from veriloggen import *

def mkTest():
    m = Module('test')
    clk = m.Reg('CLK')
    rst = m.Reg('RST')
    count = m.Reg('count', width=32)

    m.Initial(
        Systask('dumpfile', 'uut.vcd'),
        Systask('dumpvars', 0, clk, rst, count),
    )
    
    m.Initial(
        clk(0),
        Forever(clk(Not(clk), ldelay=5)) # forever #5 CLK = ~CLK;
    )

    m.Initial(
        rst(0),
        Delay(100),
        rst(1),
        Delay(100),
        rst(0),
        Delay(1000),

        count(0),
        
        While(count < 1024)(
            count( count + 1 ),
            Event(Posedge(clk))
        ),
        
        Systask('finish'),
    )

    return m

if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('')
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

from veriloggen import *

prim_mux = None

def mkPrimMux():
    global prim_mux
    if prim_mux is not None:
        return prim_mux
    
    m = Module('prim_mux')
    width = m.Parameter('WIDTH', 1)
    sel = m.Input('sel')
    ina = m.Input('ina', width=width)
    inb = m.Input('inb', width=width)
    out = m.Output('out', width=width)
    m.Assign( out(Mux(sel, ina, inb)) )
    prim_mux = m
    return m

def Pmux(m, cond, true_value, false_value, width=32):
    mux = mkPrimMux()
    sel = m.TmpWire()
    ina = m.TmpWire(width=width)
    inb = m.TmpWire(width=width)
    out = m.TmpWire(width=width)
    m.Assign(sel(cond))
    m.Assign(ina(true_value))
    m.Assign(inb(false_value))
    m.Instance(mux, '_'.join(['inst', mux.name, out.name]),
               params=[('WIDTH', width)],
               ports=[sel, ina, inb, out])
    return out
    
def mkLed():
    m = Module('blinkled')
    width = m.Parameter('WIDTH', 8)
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.OutputReg('LED', width)
    count = m.Reg('count', 32)

    m.Always(Posedge(clk))(
        If(rst)(
            count(0)
        ).Else(
            count( Pmux(m, count==1023, 0, count + 1, width=32) )
        ))
    
    m.Always(Posedge(clk))(
        If(rst)(
            led(0)
        ).Else(
            led( Pmux(m, count==1024-1, led+1, led, width=width) )
        ))
    
    return m

if __name__ == '__main__':
    led = mkLed()
    verilog = led.to_verilog('')
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import collections
import veriloggen.core.module as module
import veriloggen.core.vtypes as vtypes
import veriloggen.verilog.from_verilog as from_verilog
from veriloggen.core.collect_visitor import CollectVisitor


class Submodule(vtypes.VeriloggenNode):
    """ Sub-module wrapper """

    def __init__(self, parent, child,
                 name=None, prefix=None,
                 arg_params=None, arg_ports=None,
                 as_io=None, as_wire=None, topmodule=None):

        self.parent = parent

        # child module definition as Verilog HDL source code text
        if isinstance(child, str):
            modules = from_verilog.read_verilog_module_str(child)
            if topmodule is None:
                child = list(modules.values())[0]
            else:
                child = modules[topmodule]

        self.child = child

        if name is None:
            name = child.name

        self.name = name

        if prefix is None:
            prefix = '_'.join((name, ''))

        self.prefix = prefix

        child_params = child.get_params()
        child_ports = child.get_ports()

        if arg_params is None:
            arg_params = []
        elif not isinstance(arg_params, (dict, list, tuple)):
            arg_params = [arg_params]

        if (isinstance(arg_params, (tuple, list)) and
            len(arg_params) >= 1 and
            (not isinstance(arg_params[0], (tuple, list)) or
             len(arg_params[0]) == 1)):  # nonamed argument
            arg_params = [(cport, pport)
                          for cport, pport in zip(child_params.keys(), arg_params)]

        if not isinstance(arg_params, dict):
            arg_params = collections.OrderedDict(arg_params)

        if arg_ports is None:
            arg_ports = []
        elif not isinstance(arg_ports, (dict, list, tuple)):
            arg_ports = [arg_ports]

        if (isinstance(arg_ports, (tuple, list)) and
            len(arg_ports) >= 1 and
            (not isinstance(arg_ports[0], (tuple, list)) or
             len(arg_ports[0]) == 1)):  # nonamed argument
            arg_ports = [(cport, pport)
                         for cport, pport in zip(child_ports.keys(), arg_ports)]

        if not isinstance(arg_ports, dict):
            arg_ports = collections.OrderedDict(arg_ports)

        if as_io is None:
            as_io = ()

        if not isinstance(as_io, (tuple, list)):
            as_io = (as_io,)

        if as_wire is None:
            as_wire = ()

        if not isinstance(as_wire, (tuple, list)):
            as_wire = (as_wire,)

        as_io = [v.name if not isinstance(v, str) else
                 v for v in as_io]
        as_wire = [v.name if not isinstance(v, str) else
                   v for v in as_wire]

        # collect used parameters and localparameters
        collect_visitor = CollectVisitor()

        for port in child_ports.values():
            collect_visitor.visit(port)

        used_params = collect_visitor.names

        # params
        new_params = collections.OrderedDict()

        if arg_params:
            new_params.update(
                parent.copy_params_as_localparams(child, self.prefix,
                                                  include=arg_params.keys(),
                                                  use_fullmatch=True))

        # localparams
        # used ones only
        new_localparams = collections.OrderedDict()
        if used_params:
            new_localparams.update(
                parent.copy_params_as_localparams(child, self.prefix,
                                                  include=used_params,
                                                  exclude=arg_params.keys(),
                                                  use_fullmatch=True))
            new_localparams.update(
                parent.copy_localparams(child, self.prefix, include=used_params,
                                        use_fullmatch=True))

        # overwrite the parameter value by parameter arg
        for key, param in arg_params.items():
            new_key = ''.join([self.prefix, key])
            new_param = new_params[new_key]
            new_param.value = param

        # ports
        new_ports = collections.OrderedDict()
        if as_io:
            new_ports.update(
                parent.copy_ports(child, self.prefix, include=as_io,
                                  use_fullmatch=True))

        exclude = []
        exclude.extend(arg_ports.keys())
        exclude.extend(as_io)
        exclude.extend(as_wire)

        new_ports.update(
            parent.copy_ports_as_vars(child, self.prefix, exclude=exclude,
                                      use_fullmatch=True))

        if as_wire:
            new_ports.update(
                parent.copy_ports_as_vars(child, self.prefix, include=as_wire,
                                          use_wire=True, use_fullmatch=True))

        # for instance args
        self.all_params = collections.OrderedDict()
        self.all_raw_params = collections.OrderedDict()

        for key in child_params.keys():
            new_key = ''.join((self.prefix, key))
            if new_key in new_params:
                self.all_params[key] = new_params[new_key]
                self.all_raw_params[new_key] = new_params[new_key]

        self.all_ports = collections.OrderedDict()
        self.all_raw_ports = collections.OrderedDict()

        for key in child_ports.keys():
            new_key = ''.join((self.prefix, key))
            if key in arg_ports:
                self.all_ports[key] = arg_ports[key]
                self.all_raw_ports[new_key] = arg_ports[key]
            elif new_key in new_ports:
                self.all_ports[key] = new_ports[new_key]
                self.all_raw_ports[new_key] = new_ports[new_key]

        # instance
        self.inst = parent.Instance(child, self.name,
                                    self.all_params, self.all_ports)

    def get_raw_inst_params(self):
        return self.all_raw_params

    def get_raw_inst_ports(self):
        return self.all_raw_ports

    def get_inst_params(self):
        return self.all_params

    def get_inst_ports(self):
        return self.all_ports

    def __getitem__(self, key):
        if key in self.all_params:
            return self.all_params[key]

        if key in self.all_ports:
            return self.all_ports[key]

        raise KeyError("'%s' submodule has no item '%s'" % (self.name, key))



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *


def mkLed():
    m = Module('blinkled')
    interval = m.Parameter('INTERVAL', 8)
    length = m.Parameter('LENGTH', 8)

    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.OutputReg('LED', 8, initval=0)
    count = m.Reg('count', 32, initval=0)

    array = m.Reg('array', 32, length, initval=0)

    seq = Seq(m, 'seq', clk, rst, as_module=True)

    seq(
        Systask('display', 'LED:%d count:%d', led, count)
    )

    seq.If(count < interval - 1)(
        count(count + 1 + array[count]),
        # array[count](count) # not supported
    )
    seq.If(count == interval - 1)(
        count(0)
    )
    seq.If(count == interval - 1)(
        led(led + 1)
    )

    return m


def mkTest():
    m = Module('test')

    # target instance
    led = mkLed()

    # copy paras and ports
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)

    clk = ports['CLK']
    rst = ports['RST']

    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))

    #simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000),
        Systask('finish'),
    )

    return m


if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

def mkSubmod():
    m = Module('submod')
    pos = m.Parameter('POS', 0)
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.Output('LED')
    count = m.Reg('count', 32)
    m.Always(Posedge(clk))(
        If(rst)(
            count(0)
        ).Else(
            If(count == 1023)(
                count(0)
            ).Else(
                count(count + 1)
            )
        ))
    m.Assign( led(count[pos]) )
    return m

def mkLed():
    m = Module('blinkled')
    width = m.Parameter('WIDTH', 8)
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.Output('LED', width)

    # genvar i;
    i = m.Genvar('i')
    
    # generate for(i=0; i<WIDTH; i=i+1) begin: gen_for;
    gen_for = m.GenerateFor(i(0), i<width, i(i+1), scope='gen_for')
    submod = mkSubmod()
    params = [ ('POS', i+2) ]
    ports = [ ('CLK', clk), ('RST', rst), ('LED', led[i]) ]
    gen_for.Instance(submod, 'inst_submod', params, ports)
    # ... end endgenerate
    
    return m

if __name__ == '__main__':
    led = mkLed()
    verilog = led.to_verilog()
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

from veriloggen import *


def mkLed():
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.OutputReg('LED', 8, initval=0)

    code = m.EmbeddedCode("""
// Embedded code
reg [31:0] count;
always @(posedge CLK) begin
  if(RST) begin
    count <= 0;
    LED <= 0;
  end else begin
    if(count == 1024 - 1) begin
      count <= 0;
      LED <= LED + 1;
    end else begin
      count <= count + 1;
    end 
  end
end
""")

    return m

if __name__ == '__main__':
    led = mkLed()
    verilog = led.to_verilog()
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import numpy as np

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi
import veriloggen.types.ipcore as ipcore

axi_datawidth = 32
datawidth = 32

matrix_size = 16
a_offset = 0
b_offset = 4096
c_offset = 4096 * 2


def mkLed():
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.OutputReg('led', 8, initval=0)

    addrwidth = 10
    ram_a = vthread.RAM(m, 'ram_a', clk, rst, datawidth, addrwidth)
    ram_b = vthread.RAM(m, 'ram_b', clk, rst, datawidth, addrwidth)
    ram_c = vthread.RAM(m, 'ram_c', clk, rst, datawidth, addrwidth)

    maxi = vthread.AXIM(m, 'maxi', clk, rst, datawidth)
    saxi = vthread.AXISLiteRegister(m, 'saxi', clk, rst, datawidth, length=8)

    def matmul():
        while True:
            saxi.wait_flag(0, value=1, resetvalue=0)
            matrix_size = saxi.read(1)
            a_offset = saxi.read(2)
            b_offset = saxi.read(3)
            c_offset = saxi.read(4)
            comp(matrix_size, a_offset, b_offset, c_offset)
            saxi.write_flag(5, 1, resetvalue=0)

    def comp(matrix_size, a_offset, b_offset, c_offset):
        a_addr, c_addr = a_offset, c_offset

        for i in range(matrix_size):
            maxi.dma_read(ram_a, 0, a_addr, matrix_size)

            b_addr = b_offset
            for j in range(matrix_size):
                maxi.dma_read(ram_b, 0, b_addr, matrix_size)

                sum = 0
                for k in range(matrix_size):
                    x = ram_a.read(k)
                    y = ram_b.read(k)
                    sum += x * y
                ram_c.write(j, sum)

                b_addr += matrix_size * (datawidth // 8)

            maxi.dma_write(ram_c, 0, c_addr, matrix_size)
            a_addr += matrix_size * (datawidth // 8)
            c_addr += matrix_size * (datawidth // 8)

    th = vthread.Thread(m, 'th_matmul', clk, rst, matmul)
    fsm = th.start()

    return m


def mkTest(memimg_name=None):

    a_shape = (matrix_size, matrix_size)
    b_shape = (matrix_size, matrix_size)
    c_shape = (a_shape[0], b_shape[0])

    n_raw_a = axi.shape_to_length(a_shape)
    n_raw_b = axi.shape_to_length(b_shape)

    n_a = axi.shape_to_memory_size(a_shape, datawidth)
    n_b = axi.shape_to_memory_size(b_shape, datawidth)

    a = np.zeros(a_shape, dtype=np.int64)
    b = np.zeros(b_shape, dtype=np.int64)

    value = 1
    for y in range(a_shape[0]):
        for x in range(a_shape[1]):
            if x == y:
                a[y][x] = value
                value += 1
            else:
                a[y][x] = 0

    for y in range(b_shape[0]):
        for x in range(b_shape[1]):
            if x == y:
                b[y][x] = 2
            else:
                b[y][x] = 0

    a_addr = a_offset
    size_a = n_a * datawidth // 8
    b_addr = b_offset
    size_b = n_b * datawidth // 8

    mem = np.zeros([1024 * 1024 * 8 // axi_datawidth], dtype=np.int64)
    axi.set_memory(mem, a, axi_datawidth, datawidth, a_addr)
    axi.set_memory(mem, b, axi_datawidth, datawidth, b_addr)

    led = mkLed()

    m = Module('test')
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)
    clk = ports['CLK']
    rst = ports['RST']

    memory = axi.AxiMemoryModel(m, 'memory', clk, rst,
                                mem_datawidth=axi_datawidth,
                                memimg=mem, memimg_name=memimg_name)

    memory.connect(ports, 'maxi')

    # AXI-Slave controller
    _saxi = vthread.AXIMLite(m, '_saxi', clk, rst, noio=True)
    _saxi.connect(ports, 'saxi')

    # Timer
    counter = m.Reg('counter', 32, initval=0)
    seq = Seq(m, 'seq', clk, rst)
    seq(
        counter.inc()
    )

    def ctrl():
        for i in range(100):
            pass

        awaddr = 4
        print('# matrix_size = %d' % matrix_size)
        _saxi.write(awaddr, matrix_size)

        awaddr = 8
        print('# a_offset = %d' % a_offset)
        _saxi.write(awaddr, a_offset)

        awaddr = 12
        print('# b_offset = %d' % b_offset)
        _saxi.write(awaddr, b_offset)

        awaddr = 16
        print('# c_offset = %d' % c_offset)
        _saxi.write(awaddr, c_offset)

        awaddr = 0
        start_time = counter
        print('# start time = %d' % start_time)
        _saxi.write(awaddr, 1)

        araddr = 20
        v = _saxi.read(araddr)
        while v == 0:
            v = _saxi.read(araddr)

        end_time = counter
        print('# end time = %d' % end_time)
        time = end_time - start_time
        print('# exec time = %d' % time)

        all_ok = True
        for y in range(matrix_size):
            for x in range(matrix_size):
                v = memory.read(
                    c_offset + (y * matrix_size + x) * datawidth // 8)
                if y == x and vthread.verilog.NotEql(v, (y + 1) * 2):
                    all_ok = False
                    print("NG [%d,%d] = %d" % (y, x, v))
                if y != x and vthread.verilog.NotEql(v, 0):
                    all_ok = False
                    print("NG [%d,%d] = %d" % (y, x, v))

        if all_ok:
            print('# verify: PASSED')
        else:
            print('# verify: FAILED')

        vthread.finish()

    th = vthread.Thread(m, 'th_ctrl', clk, rst, ctrl)
    fsm = th.start()

    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))

    simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000000),
        Systask('finish'),
    )

    return m


def run(filename='tmp.v', simtype='iverilog', outputfile=None):

    if outputfile is None:
        outputfile = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + outputfile

    test = mkTest(memimg_name=memimg_name)

    if filename is not None:
        test.to_verilog(filename)

    sim = simulation.Simulator(test, sim=simtype)
    rslt = sim.run(outputfile=outputfile)
    lines = rslt.splitlines()
    if simtype == 'verilator' and lines[-1].startswith('-'):
        rslt = '\n'.join(lines[:-1])
    return rslt


if __name__ == '__main__':
    rslt = run(filename='tmp.v')
    print(rslt)

    memname = 'memimg_thread_matmul_ipcore.out'
    simcode = """
reg [31:0] counter;
always @(posedge sim_clk) begin
  if(!sim_resetn) begin
    counter <= 0;
  end else begin
    counter <= counter + 1;
  end
end

reg [31:0] _start_time;
reg [31:0] _end_time;
reg [31:0] _time;

reg [31:0] _addr;
reg [31:0] _data;
initial begin
  #1000;
  _addr = 4;
  _data = {matrix_size};
  $display("# matrix_size = %d", _data);
  slave_write_ipgen_slave_lite_memory_saxi_1(_data, _addr);

  _addr = 8;
  _data = {a_offset};
  $display("# a_offset = %d", _data);
  slave_write_ipgen_slave_lite_memory_saxi_1(_data, _addr);

  _addr = 12;
  _data = {b_offset};
  $display("# b_offset = %d", _data);
  slave_write_ipgen_slave_lite_memory_saxi_1(_data, _addr);

  _addr = 16;
  _data = {c_offset};
  $display("# c_offset = %d", _data);
  slave_write_ipgen_slave_lite_memory_saxi_1(_data, _addr);

  _addr = 0;
  _data = 1;
  _start_time = counter;
  $display("# start time = %d", _start_time);
  slave_write_ipgen_slave_lite_memory_saxi_1(_data, _addr);

  _addr = 20;
  _data = 0;
  while(_data == 0) begin
    slave_read_ipgen_slave_lite_memory_saxi_1(_data, _addr);
    nclk();
  end
  _end_time = counter;
  $display("# end time = %d", _end_time);
  _time = _end_time - _start_time;
  $display("# exec time = %d", _time);

  #10000;
  $finish;
end
""".format(matrix_size=matrix_size, a_offset=a_offset,
           b_offset=b_offset, c_offset=c_offset)

    m = mkLed()
    ipcore.to_ipcore(m, simcode=simcode, simmemimg=memname, iftype='axi')



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import collections

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

verilog_blinkled = """
module blinkled #
(
  parameter WIDTH = 8
)
(
  input CLK,
  input RST,
  output reg [WIDTH-1:0] LED,
  output [WIDTH-1:0] dummy_out0,
  output [WIDTH-1:0] dummy_out1,
  output [WIDTH-1:0] dummy_out2,
  input [WIDTH-1:0] dummy_in0,
  input [WIDTH-1:0] dummy_in1,
  input [WIDTH-1:0] dummy_in2
);

  reg [32-1:0] count;

  always @(posedge CLK) begin
    if(RST) begin
      count <= 0;
    end else begin
      if(count == 1023) begin
        count <= 0;
      end else begin
        count <= count + 1;
      end
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      LED <= 0;
    end else begin
      if(count == 1023) begin
        LED <= LED + 1;
      end 
    end
  end


endmodule
"""


def mkTop():
    m = Module('top')
    width = m.Parameter('WIDTH', 8)
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.Output('LED', width)

    sub = Submodule(m, verilog_blinkled, 'inst_blinkled',
                    arg_params=(('WIDTH', width),),
                    arg_ports=(('LED', led), ('CLK', clk), ('RST', rst)),
                    as_io=('dummy_out0', 'dummy_in0'), as_wire=('dummy_out1', 'dummy_in1'))

    return m

if __name__ == '__main__':
    top = mkTop()
    verilog = top.to_verilog('tmp.v')
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

def mkLed():
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')
    x = m.Input('x', 32)
    y = m.OutputReg('y', initval=0)
    
    seq = Seq(m, 'seq', clk, rst)
    
    seq.If(x<10)(
        y(0)
    )
    seq.If(x>=10)(
        y(1)
    )
    seq.If(x>100)(
        y(0)
    )
    
    seq.make_always()

    return m

def mkTest():
    m = Module('test')

    # target instance
    led = mkLed()
    
    # copy paras and ports
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)

    clk = ports['CLK']
    rst = ports['RST']
    x = ports['x']
    y = ports['y']
    
    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))

    reset_stmt = []
    reset_stmt.append( x(0) )
    
    simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, reset_stmt, period=100)

    nclk = simulation.next_clock
    
    init.add(
        Delay(1000),
        nclk(clk),
        x(9),
        nclk(clk),
        x(10),
        nclk(clk),
        x(101),
        Delay(1000),
        Systask('finish'),
    )

    return m
    
if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import stub__class

expected_verilog = """
module top #
  (
   parameter WIDTH = 8
  )
  (
   input CLK, 
   input RST, 
   output [WIDTH-1:0] LED
  );
  blinkled #
  (
   WIDTH
  )
  inst_blinkled
  (
   CLK,
   RST,
   LED
  );
endmodule
"""

def test():
    veriloggen.reset()
    test_module = stub__class.mkTop()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import from_verilog_pycoram_object

expected_verilog = """
module CoramMemoryBE1P #
(
  parameter CORAM_THREAD_NAME = "undefined",
  parameter CORAM_THREAD_ID = 0,
  parameter CORAM_ID = 0,
  parameter CORAM_SUB_ID = 0,
  parameter CORAM_ADDR_LEN = 10,
  parameter CORAM_DATA_WIDTH = 32
)
(
  input CLK,
  input [CORAM_ADDR_LEN - 1:0] ADDR,
  input [CORAM_DATA_WIDTH - 1:0] D,
  input WE,
  input [CORAM_MASK_WIDTH - 1:0] MASK,
  output [CORAM_DATA_WIDTH - 1:0] Q
);

  localparam CORAM_MEM_SIZE = (2 ** CORAM_ADDR_LEN);
  localparam CORAM_MASK_WIDTH = (CORAM_DATA_WIDTH / 8);

endmodule

module CoramMemory2P #
(
  parameter CORAM_THREAD_NAME = "undefined",
  parameter CORAM_THREAD_ID = 0,
  parameter CORAM_ID = 0,
  parameter CORAM_SUB_ID = 0,
  parameter CORAM_ADDR_LEN = 10,
  parameter CORAM_DATA_WIDTH = 32
)
(
  input CLK,
  input [CORAM_ADDR_LEN - 1:0] ADDR0,
  input [CORAM_DATA_WIDTH - 1:0] D0,
  input WE0,
  output [CORAM_DATA_WIDTH - 1:0] Q0,
  input [CORAM_ADDR_LEN - 1:0] ADDR1,
  input [CORAM_DATA_WIDTH - 1:0] D1,
  input WE1,
  output [CORAM_DATA_WIDTH - 1:0] Q1
);

  localparam CORAM_MEM_SIZE = (2 ** CORAM_ADDR_LEN);

endmodule

module CoramMemoryBE2P #
(
  parameter CORAM_THREAD_NAME = "undefined",
  parameter CORAM_THREAD_ID = 0,
  parameter CORAM_ID = 0,
  parameter CORAM_SUB_ID = 0,
  parameter CORAM_ADDR_LEN = 10,
  parameter CORAM_DATA_WIDTH = 32
)
(
  input CLK,
  input [CORAM_ADDR_LEN - 1:0] ADDR0,
  input [CORAM_DATA_WIDTH - 1:0] D0,
  input WE0,
  input [CORAM_MASK_WIDTH - 1:0] MASK0,
  output [CORAM_DATA_WIDTH - 1:0] Q0,
  input [CORAM_ADDR_LEN - 1:0] ADDR1,
  input [CORAM_DATA_WIDTH - 1:0] D1,
  input WE1,
  input [CORAM_MASK_WIDTH - 1:0] MASK1,
  output [CORAM_DATA_WIDTH - 1:0] Q1
);

  localparam CORAM_MEM_SIZE = (2 ** CORAM_ADDR_LEN);
  localparam CORAM_MASK_WIDTH = (CORAM_DATA_WIDTH / 8);

endmodule

module CoramInStream #
(
  parameter CORAM_THREAD_NAME = "undefined",
  parameter CORAM_THREAD_ID = 0,
  parameter CORAM_ID = 0,
  parameter CORAM_SUB_ID = 0,
  parameter CORAM_ADDR_LEN = 10,
  parameter CORAM_DATA_WIDTH = 32
)
(
  input CLK,
  input RST,
  output [CORAM_DATA_WIDTH - 1:0] Q,
  input DEQ,
  output EMPTY,
  output ALM_EMPTY
);

  localparam CORAM_MEM_SIZE = (2 ** CORAM_ADDR_LEN);

endmodule

module CoramOutStream #
(
  parameter CORAM_THREAD_NAME = "undefined",
  parameter CORAM_THREAD_ID = 0,
  parameter CORAM_ID = 0,
  parameter CORAM_SUB_ID = 0,
  parameter CORAM_ADDR_LEN = 4,
  parameter CORAM_DATA_WIDTH = 32
)
(
  input CLK,
  input RST,
  input [CORAM_DATA_WIDTH - 1:0] D,
  input ENQ,
  output FULL,
  output ALM_FULL
);

  localparam CORAM_MEM_SIZE = (2 ** CORAM_ADDR_LEN);

endmodule

module CoramRegister #
(
  parameter CORAM_THREAD_NAME = "undefined",
  parameter CORAM_THREAD_ID = 0,
  parameter CORAM_ID = 0,
  parameter CORAM_SUB_ID = 0,
  parameter CORAM_ADDR_LEN = 10,
  parameter CORAM_DATA_WIDTH = 32
)
(
  input CLK,
  input [CORAM_DATA_WIDTH - 1:0] D,
  input WE,
  output [CORAM_DATA_WIDTH - 1:0] Q
);

  localparam CORAM_MEM_SIZE = (2 ** CORAM_ADDR_LEN);

endmodule

module CoramSlaveStream #
(
  parameter CORAM_THREAD_NAME = "undefined",
  parameter CORAM_THREAD_ID = 0,
  parameter CORAM_ID = 0,
  parameter CORAM_SUB_ID = 0,
  parameter CORAM_ADDR_LEN = 4,
  parameter CORAM_DATA_WIDTH = 32
)
(
  input CLK,
  input RST,
  input [CORAM_DATA_WIDTH - 1:0] D,
  input ENQ,
  output FULL,
  output ALM_FULL,
  output [CORAM_DATA_WIDTH - 1:0] Q,
  input DEQ,
  output EMPTY,
  output ALM_EMPTY
);

  localparam CORAM_MEM_SIZE = (2 ** CORAM_ADDR_LEN);

endmodule

module userlogic #
(
  parameter W_A = 10,
  parameter W_D = 32,
  parameter W_COMM_A = 4
)
(
  input CLK,
  input RST
);

  reg [W_A - 1:0] mem_addr;
  reg [W_D - 1:0] mem_d;
  reg mem_we;
  wire [W_D - 1:0] mem_q;
  reg [W_D - 1:0] comm_d;
  reg comm_enq;
  wire comm_full;
  wire [W_D - 1:0] comm_q;
  reg comm_deq;
  wire comm_empty;

  always @(posedge CLK) begin
    if(RST) begin
      mem_addr <= 0;
      mem_d <= 0;
      mem_we <= 0;
      comm_d <= 0;
      comm_deq <= 0;
      comm_enq <= 0;
    end else begin
      mem_addr <= 0;
      mem_d <= 0;
      mem_we <= 0;
      comm_d <= 0;
      comm_deq <= 0;
      comm_enq <= 0;
    end
  end

  CoramMemory1P
  #(
    .CORAM_THREAD_NAME("ctrl_thread"),
    .CORAM_ID(0),
    .CORAM_SUB_ID(0),
    .CORAM_ADDR_LEN(W_A),
    .CORAM_DATA_WIDTH(W_D)
  )
  inst_data_memory
  (
    .CLK(CLK),
    .ADDR(mem_addr),
    .D(mem_d),
    .WE(mem_we),
    .Q(mem_q)
  );

  CoramChannel
  #(
    .CORAM_THREAD_NAME("ctrl_thread"),
    .CORAM_ID(0),
    .CORAM_ADDR_LEN(W_COMM_A),
    .CORAM_DATA_WIDTH(W_D)
  )
  inst_comm_channel
  (
    .CLK(CLK),
    .RST(RST),
    .D(comm_d),
    .ENQ(comm_enq),
    .FULL(comm_full),
    .Q(comm_q),
    .DEQ(comm_deq),
    .EMPTY(comm_empty)
  );

endmodule

module CoramMemory1P #
(
  parameter CORAM_THREAD_NAME = "undefined",
  parameter CORAM_THREAD_ID = 0,
  parameter CORAM_ID = 0,
  parameter CORAM_SUB_ID = 0,
  parameter CORAM_ADDR_LEN = 10,
  parameter CORAM_DATA_WIDTH = 32
)
(
  input CLK,
  input [CORAM_ADDR_LEN - 1:0] ADDR,
  input [CORAM_DATA_WIDTH - 1:0] D,
  input WE,
  output [CORAM_DATA_WIDTH - 1:0] Q
);

  localparam CORAM_MEM_SIZE = (2 ** CORAM_ADDR_LEN);

endmodule

module CoramChannel #
(
  parameter CORAM_THREAD_NAME = "undefined",
  parameter CORAM_THREAD_ID = 0,
  parameter CORAM_ID = 0,
  parameter CORAM_SUB_ID = 0,
  parameter CORAM_ADDR_LEN = 4,
  parameter CORAM_DATA_WIDTH = 32
)
(
  input CLK,
  input RST,
  input [CORAM_DATA_WIDTH - 1:0] D,
  input ENQ,
  output FULL,
  output ALM_FULL,
  output [CORAM_DATA_WIDTH - 1:0] Q,
  input DEQ,
  output EMPTY,
  output ALM_EMPTY
);

  localparam CORAM_MEM_SIZE = (2 ** CORAM_ADDR_LEN);

endmodule
"""

def test():
    veriloggen.reset()
    modules = from_verilog_pycoram_object.mkUserlogic()
    code = ''.join([ m.to_verilog() for m in modules.values() if not m.used ])

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi


def mkLed(axi_datawidth=32, datawidth=4, addrwidth=10):
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')

    numbanks = int(math.ceil(axi_datawidth / datawidth))
    myaxi = vthread.AXIM(m, 'myaxi', clk, rst, axi_datawidth)
    myram = vthread.MultibankRAM(m, 'myram', clk, rst, datawidth, addrwidth,
                                 numbanks=numbanks)

    saxi = vthread.AXISLiteRegister(m, 'saxi', clk, rst, 32)

    all_ok = m.TmpReg(initval=0)

    def blink(size):
        # wait start
        saxi.wait_flag(0, value=1, resetvalue=0)
        # reset done
        saxi.write(1, 0)

        all_ok.value = True

        for i in range(4):
            print('# iter %d start' % i)
            # Test for 4KB boundary check
            offset = i * 1024 * 16 + (myaxi.boundary_size - 4)
            body(size, offset)
            print('# iter %d end' % i)

        if all_ok:
            print('# verify (local): PASSED')
        else:
            print('# verify (local): FAILED')

        # result
        saxi.write(2, all_ok)

        # done
        saxi.write_flag(1, 1, resetvalue=0)

    def body(size, offset):
        # write
        for i in range(size):
            wdata = (i + 100) % (2 ** datawidth)
            myram.write(i, wdata)

        laddr = 0
        gaddr = offset
        myaxi.dma_write(myram, laddr, gaddr, size)
        print('dma_write: [%d] -> [%d]' % (laddr, gaddr))

        # write
        for i in range(size):
            wdata = (i + 1000) % (2 ** datawidth)
            myram.write(i, wdata)

        laddr = 0
        gaddr = (size + size) * 4 + offset
        myaxi.dma_write(myram, laddr, gaddr, size)
        print('dma_write: [%d] -> [%d]' % (laddr, gaddr))

        # read
        laddr = 0
        gaddr = offset
        myaxi.dma_read(myram, laddr, gaddr, size)
        print('dma_read:  [%d] <- [%d]' % (laddr, gaddr))

        for i in range(size):
            rdata = myram.read(i) & (2 ** datawidth - 1)
            verify = (i + 100) % (2 ** datawidth)
            if vthread.verilog.NotEql(rdata, verify):
                print('rdata[%d] = %d (!= %d)' % (i, rdata, verify))
                all_ok.value = False

        # read
        laddr = 0
        gaddr = (size + size) * 4 + offset
        myaxi.dma_read(myram, laddr, gaddr, size)
        print('dma_read:  [%d] <- [%d]' % (laddr, gaddr))

        for i in range(size):
            rdata = myram.read(i) & (2 ** datawidth - 1)
            verify = (i + 1000) % (2 ** datawidth)
            if vthread.verilog.NotEql(rdata, verify):
                print('rdata[%d] = %d (!= %d)' % (i, rdata, verify))
                all_ok.value = False

    th = vthread.Thread(m, 'th_blink', clk, rst, blink)
    fsm = th.start(16)

    return m


def mkTest(memimg_name=None, axi_datawidth=32, datawidth=4, addrwidth=10):
    m = Module('test')

    # target instance
    led = mkLed(axi_datawidth, datawidth, addrwidth)

    # copy paras and ports
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)

    clk = ports['CLK']
    rst = ports['RST']

    memory = axi.AxiMemoryModel(m, 'memory', clk, rst, memimg_name=memimg_name)
    memory.connect(ports, 'myaxi')

    # AXI-Slave controller
    _saxi = vthread.AXIMLite(m, '_saxi', clk, rst, noio=True)
    _saxi.connect(ports, 'saxi')

    def ctrl():
        for i in range(100):
            pass

        for i in range(16):
            # word addressing
            r = memory.read_word(i, 0, datawidth)
            print('read:  mem[%d] -> %x' % (i, r))

            # word addressing
            w = (r + i + 100) % (2 ** datawidth - 1)
            memory.write_word(i, 0, w, datawidth)
            print('write: mem[%d] <- %x' % (i, w))

        awaddr = 0
        _saxi.write(awaddr, 1)

        araddr = 4
        v = _saxi.read(araddr)
        while v == 0:
            v = _saxi.read(araddr)

        araddr = 8
        v = _saxi.read(araddr)
        if v:
            print('# verify: PASSED')
        else:
            print('# verify: FAILED')

    th = vthread.Thread(m, 'th_ctrl', clk, rst, ctrl)
    fsm = th.start()

    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))

    # simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000000),
        Systask('finish'),
    )

    return m


def run(filename='tmp.v', simtype='iverilog', outputfile=None):

    if outputfile is None:
        outputfile = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + outputfile

    test = mkTest(memimg_name=memimg_name)

    if filename is not None:
        test.to_verilog(filename)

    sim = simulation.Simulator(test, sim=simtype)
    rslt = sim.run(outputfile=outputfile)
    lines = rslt.splitlines()
    if simtype == 'verilator' and lines[-1].startswith('-'):
        rslt = '\n'.join(lines[:-1])
    return rslt


if __name__ == '__main__':
    rslt = run(filename='tmp.v')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

def mkLed():
    m = Module('blinkled')
    width = m.Parameter('WIDTH', 8)
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.OutputReg('LED', width, initval=0)

    fsm = FSM(m, 'fsm', clk, rst)
    init = fsm.current

    tmp = []
    for i in range(4):
        tmp.append( m.Reg('tmp_' + str(i), width, initval=0) )
        
    for i in range(4):
        fsm( tmp[i](fsm.current) ) 
        fsm.goto_next(cond=None)
        
    fsm( led(led + 1) )
    fsm.goto(init, cond=None)
    
    fsm.make_always()
    
    return m

if __name__ == '__main__':
    led = mkLed()
    verilog = led.to_verilog()
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *
from veriloggen.thread import RAM
import veriloggen.dataflow as dataflow


def mkMain(n=128, datawidth=32, numports=2):
    m = Module('main')

    clk = m.Input('CLK')
    rst = m.Input('RST')

    addrwidth = int(math.log(n, 2)) * 2
    myram = RAM(m, 'myram', clk, rst, datawidth, addrwidth, 2)
    myram.disable_write(1)

    df = dataflow.DataflowManager(m, clk, rst)
    fsm = FSM(m, 'fsm', clk, rst)

    # dataflow
    value = df.Counter()

    # write dataflow (Dataflow -> RAM)
    wport = 0
    waddr = 0
    wlen = 64
    done = myram.write_dataflow(wport, waddr, value, wlen, cond=fsm)
    fsm.goto_next()
    fsm.If(done).goto_next()

    fsm.goto_next()

    # read dataflow (RAM -> Dataflow)
    rport = 1
    raddr = 0
    rlen = 32
    reuse_size = 4

    rdata0, rdata1, rlast, done = myram.read_dataflow_reuse(rport, raddr, rlen,
                                                            num_outputs=2,
                                                            reuse_size=reuse_size,
                                                            cond=fsm)
    fsm.goto_next()
    fsm.If(done).goto_next()

    # verify
    rdata0_data, rdata0_valid = rdata0.read()
    rdata1_data, rdata1_valid = rdata1.read()
    rlast_data, rlast_valid = rlast.read()

    sum0 = m.Reg('sum0', 32, initval=0)
    sum1 = m.Reg('sum1', 32, initval=0)
    expected_sum = ((raddr + raddr + rlen - 1) * rlen // 2) * reuse_size

    seq = Seq(m, 'seq', clk, rst)

    seq.If(rdata0_valid)(
        sum0.add(rdata0_data),
        Systask('display', 'rdata0_data=%d', rdata0_data)
    )
    seq.If(rdata1_valid)(
        sum1.add(rdata1_data),
        Systask('display', 'rdata1_data=%d', rdata1_data)
    )
    seq.Then().If(rlast_data == 1).Delay(1)(
        Systask('display', 'sum=%d expected_sum=%d', sum0 + sum1, expected_sum)
    )

    return m


def mkTest():
    m = Module('test')

    # target instance
    main = mkMain()

    # copy paras and ports
    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']

    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    simulation.setup_waveform(m, uut, m.get_vars())
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000 * 100),
        Systask('finish'),
    )

    return m


if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    sim = simulation.Simulator(test)
    rslt = sim.run()
    print(rslt)

    # sim.view_waveform()



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))))

from veriloggen import *
import veriloggen.types.axi as axi

import veriloggen.types.ipcore as ipcore


def mkMain():
    m = Module('main')
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.Output('LED', 32)

    myaxi = axi.AxiMaster(m, 'myaxi', clk, rst)
    myaxi.disable_write()

    fsm = FSM(m, 'fsm', clk, rst)
    sum = m.Reg('sum', 32, initval=0)
    led.assign(sum)

    # read address (1)
    araddr = 1024
    arlen = 64
    expected_sum = (araddr // 4 + araddr // 4 + arlen - 1) * arlen // 2

    ack, counter = myaxi.read_request(araddr, arlen, cond=fsm)
    fsm.If(ack).goto_next()

    # read data (1)
    data, valid, last = myaxi.read_data(counter, cond=fsm)

    fsm.If(valid)(
        sum(sum + data)
    )
    fsm.Then().If(last).goto_next()

    # read address (2)
    araddr = 1024 + 1024
    arlen = 64
    expected_sum += (araddr // 4 + araddr // 4 + arlen - 1) * arlen // 2

    ack, counter = myaxi.read_request(araddr, arlen, cond=fsm)
    fsm.If(ack).goto_next()

    # read data (2)
    data, valid, last = myaxi.read_data(counter, cond=fsm)

    fsm.If(valid)(
        sum(sum + data)
    )
    fsm.Then().If(last).goto_next()

    fsm(
        Systask('display', 'sum=%d expected_sum=%d', sum, expected_sum)
    )
    fsm.goto_next()

    fsm.make_always()

    return m


def mkTest():
    m = Module('test')

    # target instance
    main = mkMain()

    # copy paras and ports
    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']

    memory = axi.AxiMemoryModel(m, 'memory', clk, rst)
    memory.connect(ports, 'myaxi')

    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    simulation.setup_waveform(m, uut, m.get_vars())
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000 * 100),
        Systask('finish'),
    )

    return m

if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    sim = simulation.Simulator(test)
    rslt = sim.run()
    print(rslt)

    # sim.view_waveform()

    simcode = """
initial begin
  #10000;
  $finish;
end
"""

    m = mkMain()
    ipcore.to_ipcore(m, 'myipcore', simcode=simcode, iftype='axi')



from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import math
import copy
import functools
from collections import OrderedDict

import veriloggen.core.vtypes as vtypes
from veriloggen.core.module import Module
from veriloggen.seq.seq import Seq

from . import visitor
from . import stypes
from . import mul
from . import scheduler
from . import allocator
from . import graph


# ID counter for 'Stream'
_stream_counter = 0


def reset():
    global _stream_counter
    _stream_counter = 0
    stypes._object_counter = 0
    mul.reset()


def StreamManager(module, clock, reset,
                  ivalid=None, iready=None,
                  ovalid=None, oready=None,
                  aswire=True, no_hook=False):
    return Stream(module=module, clock=clock, reset=reset,
                  ivalid=ivalid, iready=iready,
                  ovalid=ovalid, oready=oready,
                  aswire=aswire, no_hook=no_hook)


class Stream(object):

    def __init__(self, *nodes, **opts):
        # ID for manager reuse and merge
        global _stream_counter
        self.object_id = _stream_counter
        _stream_counter += 1

        self.nodes = set()
        self.named_numerics = OrderedDict()

        self.add(*nodes)

        self.max_stage = 0
        self.last_input = None
        self.last_output = None

        self.module = opts['module'] if 'module' in opts else None
        self.clock = opts['clock'] if 'clock' in opts else None
        self.reset = opts['reset'] if 'reset' in opts else None

        self.ivalid = opts['ivalid'] if 'ivalid' in opts else None
        self.iready = opts['iready'] if 'iready' in opts else None
        self.ovalid = opts['ovalid'] if 'ovalid' in opts else None
        self.oready = opts['oready'] if 'oready' in opts else None

        self.aswire = opts['aswire'] if 'aswire' in opts else True
        self.dump = opts['dump'] if 'dump' in opts else False
        self.dump_base = opts['dump_base'] if 'dump_base' in opts else 10
        self.dump_mode = opts['dump_mode'] if 'dump_mode' in opts else 'all'

        self.seq = None
        self.has_control = False

        self.implemented = False

        if (self.module is not None and
                self.clock is not None and self.reset is not None):

            no_hook = opts['no_hook'] if 'no_hook' in opts else False
            if not no_hook:
                self.module.add_hook(self.implement)

            seq_name = (opts['seq_name'] if 'seq_name' in opts else
                        '_stream_seq_%d' % self.object_id)
            self.seq = Seq(self.module, seq_name, self.clock, self.reset)

        if self.dump:
            dump_enable_name = '_stream_dump_enable_%d' % self.object_id
            dump_enable = self.module.Reg(dump_enable_name, initval=0)
            dump_mask_name = '_stream_dump_mask_%d' % self.object_id
            dump_mask = self.module.Reg(dump_mask_name, initval=0)
            dump_step_name = '_stream_dump_step_%d' % self.object_id
            dump_step = self.module.Reg(dump_step_name, 32, initval=0)

            self.dump_enable = dump_enable
            self.dump_mask = dump_mask
            self.dump_step = dump_step

            if self.seq:
                self.seq.add_reset(self.dump_enable)
                self.seq.add_reset(self.dump_mask)

    # -------------------------------------------------------------------------
    def add(self, *nodes):
        self.nodes.update(set(nodes))

        for node in nodes:
            if hasattr(node, 'input_data'):
                if isinstance(node.input_data, str):
                    name = node.input_data
                else:
                    name = node.input_data.name
                self.named_numerics[name] = node

            elif hasattr(node, 'output_data'):
                if node.output_data is None:
                    continue
                if isinstance(node.output_data, str):
                    name = node.output_data
                else:
                    name = node.output_data.name
                self.named_numerics[name] = node

    # -------------------------------------------------------------------------
    def to_module(self, name, clock='CLK', reset='RST', aswire=False, seq_name=None):
        """ generate a Module definion """

        m = Module(name)
        clk = m.Input(clock)
        rst = m.Input(reset)

        m = self.implement(m, clk, rst, aswire=aswire, seq_name=seq_name)

        return m

    # -------------------------------------------------------------------------
    def implement(self, m=None, clock=None, reset=None, aswire=None, seq_name=None):
        """ implemente actual registers and operations in Verilog """

        if self.implemented:
            if m is None:
                return self.module
            raise ValueError('already implemented.')

        self.implemented = True

        if m is None:
            m = self.module

        if self.module is None:
            self.module = m

        if clock is None:
            clock = self.clock

        if reset is None:
            reset = self.reset

        if self.seq is None:
            if seq_name is None:
                seq_name = '_stream_seq_%d' % self.object_id
            seq = Seq(m, seq_name, clock, reset)
        else:
            seq = self.seq

        if aswire is None:
            aswire = self.aswire

        self.add_control(aswire=aswire)
        self.has_control = True

        # for mult and div
        m._clock = clock
        m._reset = reset

        stream_nodes = self.nodes

        input_visitor = visitor.InputVisitor()
        input_vars = set()
        for node in sorted(stream_nodes, key=lambda x: x.object_id):
            input_vars.update(input_visitor.visit(node))

        output_visitor = visitor.OutputVisitor()
        output_vars = set()
        for node in sorted(stream_nodes, key=lambda x: x.object_id):
            output_vars.update(output_visitor.visit(node))

        # add input ports
        for input_var in sorted(input_vars, key=lambda x: x.object_id):
            input_var._implement_input(m, seq, aswire)

        # schedule
        sched = scheduler.ASAPScheduler()
        sched.schedule(output_vars)

        # balance output stage depth
        max_stage = 0
        for output_var in sorted(output_vars, key=lambda x: x.object_id):
            max_stage = stypes._max(max_stage, output_var.end_stage)
        self.max_stage = max_stage

        output_vars = sched.balance_output(output_vars, max_stage)

        # get all vars
        all_visitor = visitor.AllVisitor()
        all_vars = set()
        for output_var in sorted(output_vars, key=lambda x: x.object_id):
            all_vars.update(all_visitor.visit(output_var))

        # control (valid and ready)
        if not self.has_control:
            self.add_control(aswire)

        self.implement_control(seq)

        # allocate (implement signals)
        alloc = allocator.Allocator()
        alloc.allocate(m, seq, all_vars, self.valid_list, self.senable)

        # set default module information
        for var in sorted(all_vars, key=lambda x: x.object_id):
            var._set_module(m)
            var._set_strm(self)

            if var.seq is not None:
                seq.update(var.seq)

            var._set_seq(seq)

        # add output ports
        for output_var in sorted(output_vars, key=lambda x: x.object_id):
            output_var._implement_output(m, seq, aswire)

        # save schedule result
        self.last_input = input_vars
        self.last_output = output_vars

        if self.dump:
            self.add_dump(m, seq, input_vars, output_vars, all_vars)

        return m

    def add_dump(self, m, seq, input_vars, output_vars, all_vars):
        pipeline_depth = self.pipeline_depth()
        log_pipeline_depth = max(
            int(math.ceil(math.log(pipeline_depth, 10))), 1)

        seq(
            self.dump_step(1)
        )

        for i in range(pipeline_depth + 1):
            seq.If(seq.Prev(self.dump_enable, i))(
                self.dump_step.inc()
            )

        def get_name(obj):
            if hasattr(obj, 'name'):
                return obj.name
            if isinstance(obj, vtypes._Constant):
                return obj.__class__.__name__
            raise TypeError()

        longest_name_len = 0
        for input_var in sorted(input_vars, key=lambda x: x.object_id):
            if not (self.dump_mode == 'all' or
                    self.dump_mode == 'stream' or
                    self.dump_mode == 'input' or
                    self.dump_mode == 'inout' or
                    (self.dump_mode == 'selective' and
                     hasattr(input_var, 'dump') and input_var.dump)):
                continue

            name = get_name(input_var.sig_data)
            length = len(name) + 6
            longest_name_len = max(longest_name_len, length)

        for var in sorted(all_vars, key=lambda x: (-1, x.object_id)
                          if x.end_stage is None else
                          (x.end_stage, x.object_id)):
            if not (self.dump_mode == 'all' or
                    self.dump_mode == 'stream' or
                    (self.dump_mode == 'selective' and
                     hasattr(var, 'dump') and var.dump)):
                continue

            name = get_name(var.sig_data)
            length = len(name) + 6
            longest_name_len = max(longest_name_len, length)

        for output_var in sorted(output_vars, key=lambda x: x.object_id):
            if not (self.dump_mode == 'all' or
                    self.dump_mode == 'stream' or
                    self.dump_mode == 'output' or
                    self.dump_mode == 'inout' or
                    (self.dump_mode == 'selective' and
                     hasattr(output_var, 'dump') and output_var.dump)):
                continue

            name = get_name(output_var.output_sig_data)
            length = len(name) + 6
            longest_name_len = max(longest_name_len, length)

        longest_var_len = 0
        for var in sorted(all_vars, key=lambda x: (-1, x.object_id)
                          if x.start_stage is None else
                          (x.start_stage, x.object_id)):
            bit_length = var.sig_data.bit_length()
            if bit_length is None:
                bit_length = 1
            if bit_length <= 0:
                bit_length = 1

            base = (var.dump_base if hasattr(var, 'dump_base') else
                    self.dump_base)
            length = int(math.ceil(bit_length / math.log(base, 2)))
            longest_var_len = max(longest_var_len, length)

        for input_var in sorted(input_vars, key=lambda x: x.object_id):

            base = (input_var.dump_base if hasattr(input_var, 'dump_base') else
                    self.dump_base)
            base_char = ('b' if base == 2 else
                         'o' if base == 8 else
                         'd' if base == 10 and input_var.point == 0 else
                         'f' if base == 10 and input_var.point > 0 else
                         'x')
            prefix = ('0b' if base == 2 else
                      '0o' if base == 8 else
                      '  ' if base == 10 else
                      '0x')

            fmt_list = [prefix, '%', '%d' % (longest_var_len + 1), base_char]

            if input_var not in all_vars:
                fmt_list.append(' (unused)')

            input_var.dump_fmt = ''.join(fmt_list)

        for output_var in sorted(output_vars, key=lambda x: x.object_id):

            base = (output_var.dump_base if hasattr(output_var, 'dump_base') else
                    self.dump_base)
            base_char = ('b' if base == 2 else
                         'o' if base == 8 else
                         'd' if base == 10 and output_var.point == 0 else
                         'f' if base == 10 and output_var.point > 0 else
                         'x')
            prefix = ('0b' if base == 2 else
                      '0o' if base == 8 else
                      '  ' if base == 10 else
                      '0x')

            fmt_list = [prefix, '%', '%d' % (longest_var_len + 1), base_char]

            if output_var not in all_vars:
                fmt_list.append(' (unused)')

            output_var.dump_fmt = ''.join(fmt_list)

        for var in sorted(all_vars, key=lambda x: (-1, x.object_id)
                          if x.start_stage is None else
                          (x.start_stage, x.object_id)):

            base = (var.dump_base if hasattr(var, 'dump_base') else
                    self.dump_base)
            base_char = ('b' if base == 2 else
                         'o' if base == 8 else
                         'd' if base == 10 and var.point == 0 else
                         'f' if base == 10 and var.point > 0 else
                         'x')
            prefix = ('0b' if base == 2 else
                      '0o' if base == 8 else
                      '  ' if base == 10 else
                      '0x')

            fmt_list = [prefix, '%', '%d' % (longest_var_len + 1), base_char]
            var.dump_fmt = ''.join(fmt_list)

        enables = []
        for input_var in sorted(input_vars, key=lambda x: x.object_id):
            if not (self.dump_mode == 'all' or
                    self.dump_mode == 'stream' or
                    self.dump_mode == 'input' or
                    self.dump_mode == 'inout' or
                    (self.dump_mode == 'selective' and
                     hasattr(input_var, 'dump') and input_var.dump)):
                continue

            vfmt = input_var.dump_fmt

            name = get_name(input_var.sig_data)
            name_alignment = ' ' * (longest_name_len - len(name) -
                                    len('(in) '))
            fmt = ''.join(['<', self.name, ' step:%d, ',
                           'stage:%', str(
                               log_pipeline_depth), 'd, age:%d> (in) ',
                           name_alignment, name, ' = ', vfmt])

            stage = input_var.end_stage if input_var.end_stage is not None else 0
            enable = seq.Prev(self.dump_enable, stage)
            enables.append(enable)
            age = seq.Prev(self.dump_step, stage) - 1

            if input_var.point == 0:
                sig_data = input_var.sig_data
            else:
                sig_data = vtypes.Div(vtypes.SystemTask('itor', input_var.sig_data),
                                      1.0 * (2 ** input_var.point))

            seq.If(enable, vtypes.Not(self.dump_mask))(
                vtypes.Display(fmt, self.dump_step, stage, age, sig_data)
            )

        for var in sorted(all_vars, key=lambda x: (-1, x.object_id)
                          if x.end_stage is None else
                          (x.end_stage, x.object_id)):
            if not (self.dump_mode == 'all' or
                    self.dump_mode == 'stream' or
                    (self.dump_mode == 'selective' and
                     hasattr(var, 'dump') and var.dump)):
                continue

            vfmt = var.dump_fmt

            name = get_name(var.sig_data)
            name_alignment = ' ' * (longest_name_len - len(name))
            stage = var.end_stage if var.end_stage is not None else 0

            fmt = ''.join(['<', self.name, ' step:%d, ',
                           'stage:%', str(log_pipeline_depth), 'd, age:%d> ',
                           name_alignment, name, ' = ', vfmt])

            enable = seq.Prev(self.dump_enable, stage)
            enables.append(enable)
            age = seq.Prev(self.dump_step, stage) - 1

            if var.point == 0:
                sig_data = var.sig_data
            else:
                sig_data = vtypes.Div(vtypes.SystemTask('itor', var.sig_data),
                                      1.0 * (2 ** var.point))

            seq.If(enable, vtypes.Not(self.dump_mask))(
                vtypes.Display(fmt, self.dump_step, stage, age, sig_data)
            )

        for output_var in sorted(output_vars, key=lambda x: x.object_id):
            if not (self.dump_mode == 'all' or
                    self.dump_mode == 'stream' or
                    self.dump_mode == 'output' or
                    self.dump_mode == 'inout' or
                    (self.dump_mode == 'selective' and
                     hasattr(output_var, 'dump') and output_var.dump)):
                continue

            vfmt = output_var.dump_fmt

            name = get_name(output_var.output_sig_data)
            name_alignment = ' ' * (longest_name_len - len(name) -
                                    len('(out) '))
            fmt = ''.join(['<', self.name, ' step:%d, ',
                           'stage:%', str(
                               log_pipeline_depth), 'd, age:%d> (out) ',
                           name_alignment, name, ' = ', vfmt])

            stage = output_var.end_stage if output_var.end_stage is not None else 0
            enable = seq.Prev(self.dump_enable, stage)
            enables.append(enable)
            age = seq.Prev(self.dump_step, stage) - 1

            if output_var.point == 0:
                sig_data = output_var.output_sig_data
            else:
                sig_data = vtypes.Div(vtypes.SystemTask('itor', output_var.output_sig_data),
                                      1.0 * (2 ** output_var.point))

            seq.If(enable, vtypes.Not(self.dump_mask))(
                vtypes.Display(fmt, self.dump_step, stage, age, sig_data)
            )

    # -------------------------------------------------------------------------
    def add_control(self, aswire=True):
        if self.ivalid is not None and isinstance(self.ivalid, str):
            if aswire:
                self.ivalid = self.module.Wire(self.ivalid)
            else:
                self.ivalid = self.module.Input(self.ivalid)

        if self.iready is not None and isinstance(self.iready, str):
            if aswire:
                self.iready = self.module.Wire(self.iready)
            else:
                self.iready = self.module.Output(self.iready)

        if self.ovalid is not None and isinstance(self.ovalid, str):
            if aswire:
                self.ovalid = self.module.Wire(self.ovalid)
            else:
                self.ovalid = self.module.Output(self.ovalid)

        if self.oready is not None and isinstance(self.oready, str):
            if aswire:
                self.oready = self.module.Wire(self.oready)
            else:
                self.oready = self.module.Input(self.oready)

    def implement_control(self, seq):
        self.valid_list = None

        if self.ivalid is None and self.oready is None:
            if self.ovalid is not None:
                self.ovalid.assign(1)
            if self.iready is not None:
                self.iready.assign(1)
            self.senable = None
            return

        if self.oready is None:
            self._make_valid_chain(seq)
            self.senable = None
            return

        if self.ivalid is None:
            self.iready.assign(self.oready)
            self.senable = self.oready
            return

        cond = vtypes.OrList(vtypes.Not(self.ovalid), self.oready)
        self.senable = self.module.TmpWire()
        self.senable.assign(cond)

        self._make_valid_chain(seq, self.senable)
        self.iready.assign(self.senable)

    def _make_valid_chain(self, seq, cond=None):
        self.valid_list = []
        self.valid_list.append(self.ivalid)

        name = self.ivalid.name
        prev = self.ivalid

        for i in range(self.max_stage):
            v = self.module.Reg("_{}_{}".format(name, i), initval=0)
            self.valid_list.append(v)
            seq(v(prev), cond=cond)
            prev = v

        if self.ovalid is not None:
            self.ovalid.assign(prev)

    # -------------------------------------------------------------------------
    def draw_graph(self, filename='out.png', prog='dot', rankdir='LR', approx=False):
        if self.last_output is None:
            self.to_module()

        graph.draw_graph(self.last_output, filename=filename, prog=prog,
                         rankdir=rankdir, approx=approx)

    def enable_draw_graph(self, filename='out.png', prog='dot', rankdir='LR', approx=False):
        self.module.add_hook(self.draw_graph,
                             kwargs={'filename': filename, 'prog': prog,
                                     'rankdir': rankdir, 'approx': approx})

    # -------------------------------------------------------------------------
    def get_input(self):
        if self.last_input is None:
            return OrderedDict()

        ret = OrderedDict()
        for input_var in sorted(self.last_input, key=lambda x: x.object_id):
            key = str(input_var.input_data)
            value = input_var
            ret[key] = value

        return ret

    def get_output(self):
        if self.last_output is None:
            return OrderedDict()

        ret = OrderedDict()
        for output_var in sorted(self.last_output, key=lambda x: x.object_id):
            key = str(output_var.output_data)
            value = output_var
            ret[key] = value

        return ret

    # -------------------------------------------------------------------------
    def pipeline_depth(self):
        return self.max_stage

    # -------------------------------------------------------------------------
    def __getattr__(self, attr):
        try:
            return object.__getattr__(self, attr)

        except AttributeError as e:
            if attr.startswith('__') or attr not in dir(stypes):
                raise e

            func = getattr(stypes, attr)

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                v = func(*args, **kwargs)
                if isinstance(v, (tuple, list)):
                    for item in v:
                        self._set_info(item)
                else:
                    self._set_info(v)
                return v

            return wrapper

    def _set_info(self, v):
        if isinstance(v, stypes._Numeric):
            v._set_module(self.module)
            v._set_strm(self)
            v._set_seq(self.seq)

            self.add(v)

    def get_named_numeric(self, name):
        if name not in self.named_numerics:
            raise NameError("Numeric '%s' is not defined." % name)

        return self.named_numerics[name]



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import thread_reset

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  wire [8-1:0] LED;

  blinkled
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .LED(LED)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #10000;
    $finish;
  end


endmodule



module blinkled
(
  input CLK,
  input RST,
  output [8-1:0] LED
);

  assign LED = 0;
  reg [32-1:0] th_blink;
  localparam th_blink_init = 0;
  reg signed [32-1:0] _th_blink_times_0;
  reg signed [32-1:0] _th_blink_a_1;
  reg signed [32-1:0] _th_blink_b_2;
  reg signed [32-1:0] _th_blink_c_3;
  reg [32-1:0] th_subth;
  localparam th_subth_init = 0;
  reg _th_subth_called;
  reg signed [32-1:0] _th_subth_a_4;
  reg signed [32-1:0] _th_subth_b_5;
  reg signed [32-1:0] _th_subth_c_6;
  reg signed [32-1:0] _th_subth_a_7;
  reg signed [32-1:0] _th_subth_b_8;
  reg signed [32-1:0] _th_subth_c_9;
  reg signed [32-1:0] _th_subth_i_10;
  reg signed [32-1:0] _th_subth_ret_11;
  reg signed [32-1:0] _th_subth_tmp_12_13;
  reg signed [32-1:0] _th_blink_rslt_14;
  localparam th_blink_1 = 1;
  localparam th_blink_2 = 2;
  localparam th_blink_3 = 3;
  localparam th_blink_4 = 4;
  localparam th_blink_5 = 5;
  localparam th_blink_6 = 6;
  localparam th_blink_7 = 7;
  localparam th_blink_8 = 8;
  localparam th_blink_9 = 9;
  localparam th_blink_10 = 10;
  localparam th_blink_11 = 11;
  localparam th_blink_12 = 12;
  localparam th_blink_13 = 13;

  always @(posedge CLK) begin
    if(RST) begin
      th_blink <= th_blink_init;
      _th_blink_times_0 <= 0;
      _th_blink_a_1 <= 0;
      _th_blink_b_2 <= 0;
      _th_blink_c_3 <= 0;
      _th_blink_rslt_14 <= 0;
    end else begin
      case(th_blink)
        th_blink_init: begin
          _th_blink_times_0 <= 20;
          th_blink <= th_blink_1;
        end
        th_blink_1: begin
          _th_blink_a_1 <= 100;
          _th_blink_b_2 <= 200;
          _th_blink_c_3 <= 300;
          th_blink <= th_blink_2;
        end
        th_blink_2: begin
          th_blink <= th_blink_3;
        end
        th_blink_3: begin
          $display("# subth run");
          th_blink <= th_blink_4;
        end
        th_blink_4: begin
          if(th_subth == 10) begin
            th_blink <= th_blink_5;
          end 
        end
        th_blink_5: begin
          _th_blink_rslt_14 <= _th_subth_tmp_12_13;
          th_blink <= th_blink_6;
        end
        th_blink_6: begin
          $display("# subth join: rslt=%d", _th_blink_rslt_14);
          th_blink <= th_blink_7;
        end
        th_blink_7: begin
          _th_blink_a_1 <= 100;
          _th_blink_b_2 <= 200;
          _th_blink_c_3 <= 300;
          th_blink <= th_blink_8;
        end
        th_blink_8: begin
          th_blink <= th_blink_9;
        end
        th_blink_9: begin
          $display("# subth run");
          th_blink <= th_blink_10;
        end
        th_blink_10: begin
          if(th_subth == 10) begin
            th_blink <= th_blink_11;
          end 
        end
        th_blink_11: begin
          _th_blink_rslt_14 <= _th_subth_tmp_12_13;
          th_blink <= th_blink_12;
        end
        th_blink_12: begin
          $display("# subth join: rslt=%d", _th_blink_rslt_14);
          th_blink <= th_blink_13;
        end
      endcase
    end
  end

  localparam th_subth_1 = 1;
  localparam th_subth_2 = 2;
  localparam th_subth_3 = 3;
  localparam th_subth_4 = 4;
  localparam th_subth_5 = 5;
  localparam th_subth_6 = 6;
  localparam th_subth_7 = 7;
  localparam th_subth_8 = 8;
  localparam th_subth_9 = 9;
  localparam th_subth_10 = 10;

  always @(posedge CLK) begin
    if(RST) begin
      th_subth <= th_subth_init;
      _th_subth_called <= 0;
      _th_subth_a_4 <= 0;
      _th_subth_b_5 <= 0;
      _th_subth_c_6 <= 0;
      _th_subth_a_7 <= 0;
      _th_subth_b_8 <= 0;
      _th_subth_c_9 <= 0;
      _th_subth_i_10 <= 0;
      _th_subth_ret_11 <= 0;
      _th_subth_tmp_12_13 <= 0;
    end else begin
      case(th_subth)
        th_subth_init: begin
          if(th_blink == 2) begin
            _th_subth_called <= 1;
          end 
          if(th_blink == 2) begin
            _th_subth_a_4 <= _th_blink_a_1;
          end 
          if(th_blink == 2) begin
            _th_subth_b_5 <= _th_blink_b_2;
          end 
          if(th_blink == 2) begin
            _th_subth_c_6 <= _th_blink_c_3;
          end 
          if(th_blink == 8) begin
            _th_subth_called <= 1;
          end 
          if(th_blink == 8) begin
            _th_subth_a_4 <= _th_blink_a_1;
          end 
          if(th_blink == 8) begin
            _th_subth_b_5 <= _th_blink_b_2;
          end 
          if(th_blink == 8) begin
            _th_subth_c_6 <= _th_blink_c_3;
          end 
          if(th_blink == 2) begin
            th_subth <= th_subth_1;
          end 
          if(th_blink == 8) begin
            th_subth <= th_subth_1;
          end 
        end
        th_subth_1: begin
          _th_subth_a_7 <= _th_subth_a_4;
          _th_subth_b_8 <= _th_subth_b_5;
          _th_subth_c_9 <= _th_subth_c_6;
          th_subth <= th_subth_2;
        end
        th_subth_2: begin
          $display("# subth start: %d, %d, %d", _th_subth_a_7, _th_subth_b_8, _th_subth_c_9);
          th_subth <= th_subth_3;
        end
        th_subth_3: begin
          _th_subth_i_10 <= 0;
          th_subth <= th_subth_4;
        end
        th_subth_4: begin
          if(_th_subth_i_10 < 10) begin
            th_subth <= th_subth_5;
          end else begin
            th_subth <= th_subth_7;
          end
        end
        th_subth_5: begin
          $display("# subth wait: %d", _th_subth_i_10);
          th_subth <= th_subth_6;
        end
        th_subth_6: begin
          _th_subth_i_10 <= _th_subth_i_10 + 1;
          th_subth <= th_subth_4;
        end
        th_subth_7: begin
          _th_subth_ret_11 <= _th_subth_a_7 + _th_subth_b_8 + _th_subth_c_9;
          th_subth <= th_subth_8;
        end
        th_subth_8: begin
          $display("# subth end: %d", _th_subth_ret_11);
          th_subth <= th_subth_9;
        end
        th_subth_9: begin
          _th_subth_tmp_12_13 <= _th_subth_ret_11;
          th_subth <= th_subth_10;
        end
        th_subth_10: begin
          if(th_blink == 7) begin
            _th_subth_called <= 0;
          end 
          if(th_blink == 7) begin
            th_subth <= th_subth_init;
          end 
        end
      endcase
    end
  end


endmodule
"""


def test():
    veriloggen.reset()
    test_module = thread_reset.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)
